"""Defines the ViaRegionAttributes BaseModel class.
"""

from typing import Optional

from pydantic import BaseModel

from camfi.util import Field


class ViaRegionAttributes(BaseModel):
    """Contains object annotation (region)-level metadata.

    Parameters
    ----------
    score : Optional[float]
        Score of annotation. This is only relevant for annotations which have been
        obtained automatically. Score should not be set for manual annotations.
    best_peak : Optional[int]
        Period of wingbeat in pixels.
    blur_length : Optional[float]
        Length of motion blur in pixels.
    snr : Optional[float]
        Signal-to-noise ratio of best peak.
    wb_freq_up : Optional[float]
        Wingbeat frequency estimate, assuming upward motion (and zero body-length).
    wb_freq_down : Optional[float]
        Wingbeat frequency estimate, assuming downward motion (and zero body-length).
    et_up : Optional[float]
        Corrected moth exposure time, assuming upward motion.
    et_dn : Optional[float]
        Corrected moth exposure time, assuming downward motion.
    """

    score: Optional[float] = Field(None, ge=0, le=1)
    best_peak: Optional[int] = Field(
        None, gt=0, description="period of wingbeat in pixels"
    )
    blur_length: Optional[float] = Field(
        None, gt=0.0, description="length of motion blur in pixels"
    )
    snr: Optional[float] = Field(None, description="signal to noise ratio of best peak")
    wb_freq_up: Optional[float] = Field(
        None,
        ge=0.0,
        description="wingbeat frequency estimate, assuming upward motion (and zero body-length)",
    )
    wb_freq_down: Optional[float] = Field(
        None,
        ge=0.0,
        description="wingbeat frequency estimate, assuming downward motion (and zero body-length)",
    )
    et_up: Optional[float] = Field(
        None,
        ge=0.0,
        description="corrected moth exposure time, assuming upward motion",
    )
    et_dn: Optional[float] = Field(
        None,
        ge=0.0,
        description="corrected moth exposure time, assuming downward motion",
    )



"""Implements wingbeat frequency measurement from annotated images of flying insects.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from functools import cached_property, total_ordering
from math import sqrt
from pathlib import Path
from typing import Optional, Union

from bces import bces
import numpy as np
import pandas as pd
from pydantic import (
    BaseModel,
    NonNegativeInt,
    PositiveFloat,
    PositiveInt,
    ValidationError,
    validator,
)
import sklearn.mixture
import torch
from tqdm import tqdm

from camfi.datamodel.geometry import PolylineShapeAttributes
from camfi.datamodel.via import ViaRegionAttributes, ViaMetadata, ViaProject
from camfi.util import DatetimeCorrector, Field


def autocorrelation(roi: torch.Tensor, max_pixel_period: PositiveInt) -> torch.Tensor:
    """Calculates the autocorrelation along axis 1 of roi. Will run entirely on the
    device specified by roi.device, and is optimised for running on the GPU.

    Parameters
    ----------
    roi : torch.Tensor
        Tensor of shape (width, blur_length).
    max_pixel_period : Optional[PositiveInt]
        Maximum period to consider (choosing a smaller number will increase the speed of
        execution if running on cpu, and decrease memory consumption regardless of
        which device it's running on). If None, calculated as roi.shape[1] // 2.

    Returns
    -------
    mean_autocorrelation : torch.Tensor
        Tensor of shape (max_pixel_period,). Contains the autocorrelation with of the
        roi along axis 1, with integer step-sizes.

    Examples
    --------
    >>> from torch import arange, cos, sin, stack
    >>> from math import pi
    >>> theta = arange(0., 4. * pi, pi / 4)
    >>> autocorrelation(stack([sin(theta), cos(theta)]), len(theta) // 2)
    tensor([ 0.4375,  0.2500,  0.0000, -0.2500, -0.3750, -0.2500,  0.1250,  0.3750])
    """
    mean_diff = roi - roi.mean(axis=0)  # type: ignore[call-overload]
    std = roi.std(axis=0)  # type: ignore[call-overload]

    index = torch.arange(roi.shape[1] - max_pixel_period, device=roi.device)
    step, origin = torch.meshgrid(index[: max_pixel_period + 1], index)
    step = step + origin

    autocovariance = (mean_diff[:, origin] * mean_diff[:, step]).mean(axis=0)

    denominator = std[origin] * std[step]

    return (autocovariance / denominator).nan_to_num().mean(axis=1)


def find_best_peak(
    values: torch.Tensor,
) -> tuple[Optional[PositiveInt], Optional[float]]:
    """Takes a Tensor of values (with 1 dimension), and finds the index of the best peak
    If peak finding fails, (None, None) is returned.

    Parameters
    ----------
    values : torch.Tensor
        1-D tensor of values to find peaks in.

    Returns
    -------
    best_peak : Optional[PositiveInt]
        Index of best peak.
    snr : Optional[float]
        Score of peak.

    Examples
    --------
    >>> t = torch.zeros(100)
    >>> t[0] = 1.  # The first peak is always ignored
    >>> t[50] = 1.
    >>> find_best_peak(t)
    (50, inf)

    >>> from torch import cos, linspace
    >>> from math import pi
    >>> t = (cos(linspace(0., 8. * pi, 32)) + 1) * linspace(0.5, 0., 32)
    >>> t[4] = max(t[3], t[5]) + 0.01  # small local peak is not picked up
    >>> t[2] = t[0] + 0.01  # The highest peak is not necessarily the best peak
    >>> best_peak, score = find_best_peak(t)
    >>> best_peak
    8
    >>> 0 < score < 10
    True

    If no peak is found, then (None, None) is returned
    >>> find_best_peak(torch.zeros(10))
    (None, None)
    """
    peaks = ((values > values.roll(-1)) & (values > values.roll(1))).nonzero(
        as_tuple=True
    )[0][1:]
    sorted_peaks = peaks[values[peaks].argsort(descending=True)]

    best_peak: Optional[PositiveInt] = None
    snr: Optional[float] = None
    snrs: list[float] = []

    for peak in sorted_peaks:
        peak_idx = int(peak)
        # Find snr
        trough_values = torch.cat(
            [
                values[peak_idx // 3 : (peak_idx * 3) // 4],
                values[(peak_idx * 5) // 3 : (peak_idx * 7) // 4],
            ]
        )
        snr = float((values[peak] - trough_values.mean()) / trough_values.std())
        if len(snrs) > 0:
            if snrs[-1] > snr:
                snr = snrs[-1]
                break  # Previous peak was the best peak
        snrs.append(snr)
        best_peak = int(peak)

    return best_peak, snr


class WingbeatSuppFigPlotter(ABC, BaseModel):
    """Defines an interface for producing suplementary figures from the wingbeat
    extraction procedure, used by WingbeatExtractor. This abstract base class is
    agnostic to which plotting library is used. Subclasses will naturally have to use
    a particular plotting library, such as matplotlib.

    Concrete subclasses must implement the .__call__ method.

    Parameters
    ----------
    root: Path
        Root directory to put supplementary figures in.
    image_filename: Path
        Relative path to image file. `self.get_filepath` uses this to compute the path
        where the supplementary figure will be written (note that if annotation_idx is
        set, then the filename will be modified to include the annotation index - useful
        for plotting multple supplementary figures relating to annotations from one
        image).
    suffix: str
        File suffix for output plots. Defaults to ".png", but could be ".pdf", ".html",
        or ".eps", etc. depending on the subclass of WingbeatSuppFigPlotter.
    annotation_idx: NonNegativeInt
        Used to infer the correct filename. Defaults to 0 (which is almost always what
        you would want when initialising a subclass of WingbeatSuppFigPlotter).

    Examples
    --------
    >>> class DummyWingbeatSuppFigPlotter(WingbeatSuppFigPlotter):
    ...     def __call__(
    ...         self,
    ...         region_attributes: ViaRegionAttributes,
    ...         region_of_interest: torch.Tensor,
    ...         mean_autocorrelation: torch.Tensor,
    ...     ) -> None:
    ...         self.get_filepath()
    ...         return None
    >>> supplementary_figure_plotter = DummyWingbeatSuppFigPlotter(
    ...     root="foo", image_filename="bar/baz.jpg"
    ... )
    >>> supplementary_figure_plotter.get_filepath() == Path("foo/bar/baz_0.png")
    True

    The annotation index has now been incremented by 1.

    >>> supplementary_figure_plotter.annotation_idx
    1
    >>> supplementary_figure_plotter.get_filepath() == Path("foo/bar/baz_1.png")
    True
    """

    root: Path
    image_filename: Path
    suffix: str = ".png"
    annotation_idx: NonNegativeInt = 0

    def get_filepath(self):
        """Computes the filepath for the supplementary figure and increments
        `self.annotation_idx` by 1.

        Returns
        -------
        filepath : Path
            Full path to supplementary figure file.
        """
        name = f"{self.image_filename.stem}_{self.annotation_idx}{self.suffix}"
        self.annotation_idx += 1
        return self.root.joinpath(self.image_filename.parent, name)

    @abstractmethod
    def __call__(
        self,
        region_attributes: ViaRegionAttributes,
        region_of_interest: torch.Tensor,
        mean_autocorrelation: torch.Tensor,
    ) -> None:
        """Implementations produce a supplementary figure of a wingbeat extraction
        process, perhaps writing this to a file. Should call `self.get_filepath()` once
        only to get the path to the file where the figure should be written.

        Parameters
        ----------
        region_attributes : ViaRegionAttributes
            With fields calculated (e.g. by WingbeatExtractor.process_blur)
        region_of_interest : torch.Tensor
            Greyscale image Tensor displaying region of interest
        mean_autocorrelation : torch.Tensor
            1-d Tensor with values containing autocorrrelation along axis 1 of
            `region_of_interest`
        """


class WingbeatExtractorConfig(BaseModel):
    """Contains configurable parameters for WingbeatExtractor. Provides a creation
    methods for WingbeatExtractor. See WingbeatExtractor for futher documentation.
    """

    device: str = Field(
        "cpu",
        description="Using GPU ('cuda') can give 4x speedups for certain operations.",
    )
    backup_device: Optional[str] = Field(
        None,
        description="Used when computation fails on main device due to memory limit.",
    )
    scan_distance: PositiveInt = Field(
        50,
        description="Max. distance from polyline used for wingbeat extraction.",
    )
    max_pixel_period: Optional[PositiveInt] = Field(
        None, description="Max. pixel period to check during wingbeat extraction."
    )
    force_load_exif_metadata: bool = Field(
        False,
        description="Forces EXIF metadata to be read from image instead of ViaProject.",
    )

    class Config:
        schema_extra = {
            "description": "Contains configurable parameters for WingbeatExtractor."
        }

    @validator("device", "backup_device")
    def check_device_available(cls, v):
        if v == "cuda" and not torch.cuda.is_available():
            raise ValueError(
                "No NVIDIA driver on your system. Install one or set device to 'cpu'."
            )
        try:
            torch.tensor([], device=v)
        except RuntimeError as e:
            raise ValueError(*e.args())

        return v

    def get_wingbeat_extractor(self, **kwargs):
        """Instantiates WingbeatExtractor.

        Parameters
        ----------
        **kwargs
            Passed to WingbeatExtractor

        Returns
        -------
        wingbeat_extractor : WingbeatExtractor
            WingbeatExtractor with fields taken from self and kwargs.
        """
        return WingbeatExtractor(**self.dict(), **kwargs)


class WingbeatExtractor(WingbeatExtractorConfig):
    """Class for measuring wingbeat frequencies of annotated flying insects in an image.
    A new instance of WingbeatExtractor should be used for each distinct image file.

    Parameters
    ----------
    metadata: ViaMetadata
        Containing annotations of flying insects, as well as file-level image metadata.
        If file-level metadata is missing (specifically, exposure_time), this will be
        read from the image file.
    root: Path
        Path to root directory containing all image directories.
    line_rate: PositiveFloat
        Rolling shutter line rate of the camera used to take the photo, in lines per
        second. This must be measured separately. See
        https://camfi.readthedocs.io/en/latest/usage/notebooks/camera_calibration.html
        for a guide on measuring the rolling shutter line rate.
    device: str
        Some steps can run on the GPU, which can give speedups of over 4x. To enable,
        set e.g. device="cuda".
    backup_device: Optional[str]
        If a step raises a RuntimeError, it can be re-attempted on an alternative
        device. It is recommended to set to "cpu" if running on a GPU with limited
        memory.
    scan_distance: PositiveInt
        Optional parameter used in WingbeatExtractor.process_blur. This defines the
        maximum perpendicular distance from the polyline annotation of pixels included
        in the rotated, cropped, and straightened region of interest images.
    max_pixel_period: Optional[PositiveInt]
        Optional parameter used in WingbeatExtractor.process_blur. By default,
        autocorrelation is calculated up to half the length of each motion blur. For
        speed of execution or to reduce memory footprint, a maximum value can be set.
    force_load_exif_metadata: bool
        If True, EXIF metadata will be read from the image file, regardless of whether
        exposure_time is already set in metadata.file_attributes. By default, EXIF
        metadata will only be read if it is missing from metadata.file_attributes.
    location: Optional[str]
        Sets location string when loading EXIF metadata, placed in
        metadata.file_attributes.location. Has no effect if EXIF metadata is not loaded.
        Recommended to use in conjunction with force_load_exif_metadata.
    datetime_corrector: Optional[DatetimeCorrector]
        If provided, will be called while loading EXIF metadata to obtain a corrected
        timestamp, which is placed in metadata.file_attributes.datetime_corrected. Has
        no effect if EXIF metadata is not loaded.
        Recommended to use in conjunction with force_load_exif_metadata.
    supplementary_figure_plotter: Optional[WingbeatSuppFigPlotter]
        If set, will be called to plot supplementary figures during self.process_blur.
        A custom implementation of WingbeatSuppFigPlotter may be used, or one from
        camfi.plotting.
    """

    metadata: ViaMetadata
    root: Path
    line_rate: PositiveFloat
    location: Optional[str] = None
    datetime_corrector: Optional[DatetimeCorrector] = None
    supplementary_figure_plotter: Optional[WingbeatSuppFigPlotter] = None

    # image and exposure_time may require expensive IO operations, so should only happen
    # once each, if at all. They should also be treated as immutable for the life of the
    # WingbeatExtractor instance. Hence, the property and cache decorators.
    @cached_property
    def image(self) -> torch.Tensor:
        """Loads image from file and converts it to a greyscale tensor. Output is cached
        (so image is only loaded once for the life of the WingbeatExtractor instance).

        Returns
        -------
        image : torch.Tensor
            Image tensor with shape [height, width].
        """
        return self.metadata.read_image(root=self.root).to(self.device).mean(axis=-3)  # type: ignore[call-overload]

    @cached_property
    def exposure_time(self) -> PositiveFloat:
        """Gets exposure time either from self.metadata.file_attributes, or from the
        EXIF metadata of the image file. Caches output so file is only read once for
        life of WingbeatExtractor instance.

        Returns
        -------
        expoosure_time : PositiveFloat
            Exposure time of photograph in seconds.
        """
        if (
            self.force_load_exif_metadata
            or self.metadata.file_attributes.exposure_time is None
        ):
            self.metadata.load_exif_metadata(
                root=self.root,
                location=self.location,
                datetime_corrector=self.datetime_corrector,
            )
        assert isinstance(self.metadata.file_attributes.exposure_time, float)
        return self.metadata.file_attributes.exposure_time

    class Config:
        keep_untouched = (cached_property,)

    # In order to use the cache decorator, we need to define __eq__ and __hash__
    def __eq__(self, other):
        return self is other

    def __hash__(self):
        return id(self) // 16

    def process_blur(
        self, polyline: PolylineShapeAttributes, score: Optional[float] = None
    ) -> ViaRegionAttributes:
        """Performs the camfi algorithm to takes a measurement of wingbeat frequency
        from a flying insect motion blur which has been annotated with a polyline.

        Parameters
        ----------
        polyline: PolylineShapeAttributes
            Polyline annotation following the path of the flying insect's motion blur.
        score: Optional[float]
            Score parameter to be passed to ViaRegionAttributes constructor (should set
            if processing an annotation which was generated automatically, so that the
            score is reflected in the output).

        Returns
        -------
        region_attributes : ViaRegionAttributes
            With all fields set (including score iff a value was given).
        """
        # Load region of interest. Mypy complains about self.image not being the right
        try:
            roi = polyline.extract_region_of_interest(self.image, self.scan_distance)
        except RuntimeError:
            if self.backup_device is None:
                raise
            roi = polyline.extract_region_of_interest(
                self.image.to(self.backup_device), self.scan_distance
            ).to(self.device)

        # Infer max_pixel_period if not set
        max_pixel_period = roi.shape[1] // 2
        if self.max_pixel_period is not None:
            max_pixel_period = min(max_pixel_period, self.max_pixel_period)

        # Calculate autocorrelation. This step can run on the GPU
        try:
            mean_autocorrelation = autocorrelation(roi, max_pixel_period).cpu()
        except RuntimeError:
            if self.backup_device is None:
                raise
            mean_autocorrelation = autocorrelation(
                roi.to(self.backup_device), max_pixel_period
            ).cpu()

        # Find wingbeat peak
        best_peak, snr = find_best_peak(mean_autocorrelation)

        if best_peak is None:  # Failed to find peak
            region_attributes = ViaRegionAttributes(
                score=score, blur_length=polyline.length()
            )

        else:
            # Calculate wingbeat frequency from peak.
            # Note that due to the ambiguity in the direction of the moth's flight,
            # as well as rolling shutter, there are two possible frequency estimates,
            # with the lower frequency corresponding to an upward direction and the
            # higher frequency corresponding to a downward direction of flight with
            # respect to the camera's orientation.
            y_diff = polyline.y_diff()
            corrected_exposure_time = (
                (self.exposure_time + torch.tensor([y_diff, -y_diff]) / self.line_rate)
                .sort()
                .values
            )
            period = [
                torch.arange(1, max_pixel_period + 1) * float(et) / roi.shape[1]
                for et in corrected_exposure_time
            ]
            wb_freq_up, wb_freq_down = [1 / period[i][best_peak] for i in (0, 1)]

            region_attributes = ViaRegionAttributes(
                score=score,
                best_peak=best_peak,
                blur_length=polyline.length(),
                snr=snr,
                wb_freq_up=wb_freq_up,
                wb_freq_down=wb_freq_down,
                et_up=float(corrected_exposure_time[0]),
                et_dn=float(corrected_exposure_time[1]),
            )

        # Plot supplementary figure
        if self.supplementary_figure_plotter is not None:
            self.supplementary_figure_plotter(
                region_attributes, roi, mean_autocorrelation
            )

        return region_attributes

    def extract_wingbeats(self) -> int:
        """Calls self.process_blur on the shape_attributes of all polyline regions in
        self.metadata, replacing the region_attributes of those regions with ones
        containing wingbeat data.

        Operates in place.

        Returns
        -------
        polylines_processed : int
            Number of polyline annotations processed.
        """
        polylines_processed = 0
        for region in filter(
            lambda r: r.shape_attributes.name == "polyline", self.metadata.regions
        ):
            # Get parameters for self.process_blur
            polyline = region.shape_attributes
            assert isinstance(polyline, PolylineShapeAttributes)
            score = region.region_attributes.score

            # Update region attributes with wingbeat data
            region.region_attributes = self.process_blur(polyline, score=score)
            polylines_processed += 1

        return polylines_processed


def extract_all_wingbeats(
    via_project: ViaProject, disable_progress_bar: Optional[bool] = True, **kwargs
) -> None:
    """Extracts wingbeat data from all images in via_project, inserting that data into
    via_project (operates in place).

    Parameters
    ----------
    via_project : ViaProject
        Contains metadata for each image to process.
    disable_progress_bar : Optional[bool]
        If True (default), progress bar is disabled.
        If set to None, disable on non-TTY.
    **kwargs
        Passed to WingbeatExtractor constructor (once for each image).
    """
    postfix = {"polylines_processed": 0}
    pb = tqdm(
        via_project.via_img_metadata.items(),
        disable=disable_progress_bar,
        desc="Extracting wingbeats",
        unit="img",
        dynamic_ncols=True,
        ascii=True,
        color="green",
        postfix=postfix,
    )
    for img_key, metadata in pb:
        wingbeat_extractor = WingbeatExtractor(metadata=metadata, **kwargs)
        postfix["polylines_processed"] += wingbeat_extractor.extract_wingbeats()
        via_project.via_img_metadata[img_key] = wingbeat_extractor.metadata
        pb.set_postfix(postfix, refresh=False)


@total_ordering
class BcesResult(BaseModel):
    """Stores parameters of one BCES linear regression.

    estimates : list[tuple[float, float, float, float, float]]
        list of (gradient, y_intercept, gradient_stderr, y_intercept_stderr, cov_xy)
        tuples of estimates and standard errors. Has length n_classes.
    err : np.ndarray
        Array of weighted errors for each measurement from each regression line. Has
        shape (n_classes, len(x)).

    Parameters
    ----------
    gradient : float
        Gradient Estimate.
    y_intercept : float
        Intercept estimate.
    gradient_stderr : float
        Standard error of the gradient esitmate.
    y_intercept_stderr : float
        Standard error of the y_intercept estimate.
    cov_xy : float
        Covariance estimate.
    """

    gradient: float
    y_intercept: float
    gradient_stderr: float
    y_intercept_stderr: float
    cov_xy: float

    def __lt__(self, other: BcesResult) -> bool:
        return (
            self.gradient,
            self.y_intercept,
            self.gradient_stderr,
            self.y_intercept_stderr,
            self.cov_xy,
        ) < (
            other.gradient,
            other.y_intercept,
            other.gradient_stderr,
            other.y_intercept_stderr,
            other.cov_xy,
        )


class BcesEM(BaseModel):
    """Implements an expectation-maximisation algorithm for fitting multiple BCES
    linear regresssions to a dataset.

    Parameters
    ----------
    x : np.ndarray
        Independent variable values.
    y : np.ndarray
        Dependent variable values. Should have same shape as x.
    n_classes : int
        Number of classes (i.e. number of regressions in multiple regression).
    xerr : Union[float, np.ndarray]
        Error of independent variable measurments. Should have same shape as x. If float
        is given, it will be converted to an array.
    yerr : Union[float, np.ndarray]
        Error of dependent variable measurements. Should have same shape as x. If float
        is given, it will be converted to an array.
    cov : Union[float, np.ndarray]
        Covariance of independent and dependent variable measurements. Should have same
        shape as x. If float is given, it will be converted to an array.
    class_mask: Union[None, int, np.random.Generator, np.ndarray]
        Array of integers defining classes of measurements. Accepts values
        from the set {0, 1, ..., n_classes - 1}. Should have same shape as x.
        Alternatively, give a seed for a random number generator (or the Generator
        itself), and class_mask will be generated.
    prob_class : np.ndarray
        Marginal probabilities of each class. Should have shape (n_classes,), and sum
        to 1.
    """

    x: np.ndarray
    y: np.ndarray
    n_classes: PositiveInt
    xerr: Union[float, np.ndarray] = 0.0
    yerr: Union[float, np.ndarray] = 0.0
    cov: Union[float, np.ndarray] = 0.0
    class_mask: Union[None, int, np.random.Generator, np.ndarray] = None
    prob_class: np.ndarray = None  # type: ignore[assignment]

    class Config:
        arbitrary_types_allowed = True

    @validator("x", pre=True)
    def x_is_1d(cls, v):
        assert len(v.shape) == 1
        return v

    @validator("y", "xerr", "yerr", "cov", pre=True, always=True)
    def values_same_length_as_x(cls, v, values):
        if isinstance(v, float):
            v = np.ones_like(values["x"]) * v
        if v.shape != values["x"].shape:
            raise ValidationError(
                f"Data must have same shape. {values['x'].shape}, {v.shape}."
            )
        return v

    @validator("class_mask", pre=True, always=True)
    def check_class_mask(cls, v, values):
        if v is None or isinstance(v, int):
            rng = np.random.default_rng(v)
            v = rng.integers(0, values["n_classes"], len(values["x"]), "u1")
        elif isinstance(v, np.random.Generator):
            v = v.integers(0, values["n_classes"], len(values["x"]), "u1")

        if v.shape != values["x"].shape:
            raise ValidationError(
                f"Data must have same shape. {values['x'].shape}, {v.shape}."
            )

        assert (
            v.min() >= 0
        ), f"class_mask must not contain negative values. Got {v.min()}."
        assert (
            v.max() < values["n_classes"]
        ), f"class_mask must not contain values greter than n_classes. Got {v.max()}."

        return v

    @validator("prob_class", pre=True, always=True)
    def check_prob_class(cls, v, values):
        if v is None:
            v = np.array(
                [np.mean(values["class_mask"] == i) for i in range(values["n_classes"])]
            )
        assert isinstance(v, np.ndarray), f"Expected np.ndarray, got {type(v)}."
        assert (
            len(v) == values["n_classes"]
        ), f"Expected {values['n_classes']} probabilities, got {len(v)}."
        assert v.min() >= 0.0, f"Probabilities cannot be negative. Got {v.min()}."
        assert v.max() <= 1.0, f"Probabilities cannot be greater than 1. got {v.max()}."
        assert (
            abs(1.0 - v.sum()) <= 1e-6
        ), f"Probabilities should sum to 1, but sum to {v.sum()}."
        return v

    @classmethod
    def from_region_dataframe(
        cls, regions: pd.DataFrame, n_classes: int, seed: Optional[int] = None
    ) -> BcesEM:
        """Initialises a BcesEM object from a regions dataframe.

        Parameters
        ----------
        regions : pd.DataFrame
            DataFrame containing wingbeat-extracted polylines. SNR threshold should
            already have been applied. Must contain columns "best_peak", "et_up",
            "et_dn", and "blur_length".
        n_classes : int
            Number of target classes.
        seed : Optional[int]
            Sets the seed for initialisation of class_mask.

        Returns
        -------
        bces_em : BcesEM
            Model to fit by calling bces_em.fit().
        """
        x = (
            np.array(
                regions["best_peak"] * regions["et_up"]
                + regions["best_peak"] * regions["et_up"]
            )
            / 2
        )
        xerr = (
            np.abs(
                np.array(
                    regions["best_peak"] * regions["et_up"]
                    - regions["best_peak"] * regions["et_up"]
                )
            )
            / 2
        )
        y = np.array(regions["blur_length"])
        yerr = np.zeros_like(y)
        cov = np.zeros_like(y)

        return BcesEM(
            x=x,
            y=y,
            n_classes=n_classes,
            xerr=xerr,
            yerr=yerr,
            cov=cov,
            class_mask=seed,
        )

    def fit_bces(self):
        """Fits BCES linear regressions to the data, subdivided into self.n_classes
        classes by self.class_mask.

        Returns
        -------
        estimates : list[BcesResult]
            list of self.n_classes BcesResult instances.
        err : np.ndarray
            Array of weighted errors for each measurement from each regression line. Has
            shape (n_classes, len(x)).
        """
        assert isinstance(self.class_mask, np.ndarray)
        estimates: list[BcesResult] = []
        err = np.zeros((self.n_classes, len(self.x)))
        for class_id in range(self.n_classes):
            mask = self.class_mask == class_id
            gradient, y_intercept, gradient_err, y_intercept_err, cov_xy = (
                e[0]
                for e in bces.bces(
                    self.x[mask],
                    self.xerr[mask],
                    self.y[mask],
                    self.yerr[mask],
                    self.cov[mask],
                )
            )
            estimates.append(
                BcesResult(
                    gradient=gradient,
                    y_intercept=y_intercept,
                    gradient_stderr=gradient_err,
                    y_intercept_stderr=y_intercept_err,
                    cov_xy=cov_xy,
                )
            )

            err[class_id, :] = (
                y_intercept + gradient * self.x - self.y
            ) ** 2 / self.prob_class[class_id]

        return estimates, err

    def fit(self, max_iterations: int = 100):
        """Performs expectation-maximisation to fit data to self.n_classes BCES linear
        regression lines. Modifies self.class_mask and self.prob_class.

        Parameters
        ----------
        max_iterations : int
            Maximum number of EM-iterations.

        Returns
        -------
        estimates : list[BcesResult]
            list of self.n_classes BcesResult instances.
        """
        assert isinstance(self.class_mask, np.ndarray)
        for i in range(max_iterations):
            # Fit BCES regressions based on existing class data
            estimates, err = self.fit_bces()

            # Update self.class_mask
            class_mask = np.argmin(err, axis=0)

            # Check if converged
            if (class_mask == self.class_mask).all():
                break

            # Reinitialise for next iteration
            self.class_mask[:] = class_mask
            self.prob_class[:] = np.array(
                [np.mean(self.class_mask == i) for i in range(self.n_classes)]
            )

        return estimates


@total_ordering
class WeightedGaussian(BaseModel):
    mean: float
    std: float
    weight: float = 1.0

    def __lt__(self, other: WeightedGaussian):
        return (self.mean, self.std, self.weight) < (
            other.mean,
            other.std,
            other.weight,
        )


class GMM(BaseModel):
    x: np.ndarray
    n_classes: PositiveInt
    seed: Optional[int] = None

    class Config:
        arbitrary_types_allowed = True

    @validator("x")
    def x_is_1d(cls, v):
        assert len(v.shape) == 1, "Data must be 1-D. Got array with shape {v.shape}."
        return v

    @classmethod
    def log10_from_region_dataframe(
        cls, regions: pd.DataFrame, n_classes: int, seed: Optional[int] = None
    ) -> GMM:
        """Initialises a GMM object from a regions dataframe, taking the log10 values
        of preliminary wingbeat frequency.

        Parameters
        ----------
        regions : pd.DataFrame
            DataFrame containing wingbeat-extracted polylines. SNR threshold should
            already have been applied. Must contain columns "wb_freq_down" and
            "wb_freq_up".
        n_classes : int
            Number of target classes.
        seed : Optional[int]
            Sets the seed for the EM algorithm.

        Returns
        -------
        gmm : GMM
            Model to fit by calling gmm.fit().
        """
        return GMM(
            x=np.log10(
                np.concatenate([regions["wb_freq_down"], regions["wb_freq_up"]])
            ),
            n_classes=n_classes,
            seed=seed,
        )

    def fit(self) -> list[WeightedGaussian]:
        gmm = sklearn.mixture.GaussianMixture(
            n_components=self.n_classes, random_state=self.seed
        )
        gmm.fit(self.x.reshape(-1, 1))
        components: list[WeightedGaussian] = []
        for i in range(self.n_classes):
            components.append(
                WeightedGaussian(
                    mean=gmm.means_[i][0],
                    std=sqrt(gmm.covariances_[i][0]),
                    weight=gmm.weights_[i],
                )
            )

        return components



"""Defines transformations used for data augmentation during automatic annotation model
training. Depends on camfi.datamodel.autoannotation."""

from math import inf, sqrt
import random
from typing import Iterable, Optional, Sequence, Union

from pydantic import BaseModel
import torch

from camfi.datamodel.autoannotation import ImageTransform, Target
from camfi.util import Field


class Compose(ImageTransform):
    """Composition of transformations.

    Parameters
    ----------
    transforms: Sequence[ImageTransform]
        Sequence of transorms to compose.
    """

    transforms: Sequence[ImageTransform]

    def apply_to_tensor_dict(
        self, image: torch.Tensor, target: dict[str, torch.Tensor]
    ) -> tuple[torch.Tensor, dict[str, torch.Tensor]]:
        """Applies each transormation in self.transforms to an image and target dict.

        Parameters
        ----------
        image : torch.Tensor
            Input image to first transformation.
        target : dict[str, torch.Tensor]
            Target tensor dict of annotations on image.

        Returns
        -------
        transformed_image : torch.Tensor
            Output image of last transformation.
        transformed_target : dict[str, torch.Tensor]
            Output target dict of last transformation.
        """
        for transform in self.transforms:
            image, target = transform.apply_to_tensor_dict(image, target)
        return image, target


class RandomHorizontalFlip(ImageTransform):
    """Image transform which applies a horizontal flip to an image with a fixed
    probability.

    Parameters
    ----------
    prob : float
        Probability of flipping image.
    """

    prob: float = Field(..., ge=0.0, le=1.0)

    def apply_to_tensor_dict(
        self, image: torch.Tensor, target: dict[str, torch.Tensor]
    ) -> tuple[torch.Tensor, dict[str, torch.Tensor]]:
        """Applies each random horizontal flip transormation to an image and target
        dict.

        Parameters
        ----------
        image : torch.Tensor
            Input image tensor.
        target : dict[str, torch.Tensor]
            Input target tensor dict.

        Returns
        -------
        transformed_image : torch.Tensor
            Output image of transformation.
        transformed_target : dict[str, torch.Tensor]
            Output target dict of transformation.
        """
        if random.random() < self.prob:
            height, width = image.shape[-2:]
            image = image.flip(-1)
            try:
                bbox = target["boxes"]
                bbox[:, [0, 2]] = width - bbox[:, [2, 0]]
                target["boxes"] = bbox
            except KeyError:
                pass
            try:
                target["masks"] = target["masks"].flip(-1)
            except KeyError:
                pass

        return image, target



"""Provides camfi project config parser.
"""

from __future__ import annotations

from datetime import date, timezone
from functools import cached_property
import json
from pathlib import Path
from typing import Any, Optional, Union, Sequence
from zipfile import ZipFile

import pandas as pd
from pydantic import (
    BaseModel,
    DirectoryPath,
    FilePath,
    NonNegativeInt,
    PositiveFloat,
    PositiveInt,
    ValidationError,
    root_validator,
    validator,
)
from strictyaml import as_document, load
from tqdm import tqdm

from camfi.annotator import (
    Annotator,
    AnnotationValidationResult,
    train_model,
    validate_annotations,
)
from camfi.datamodel.autoannotation import CamfiDataset, MaskMaker
from camfi.datamodel.geometry import BoundingBox
from camfi.datamodel.locationtime import LocationTimeCollector
from camfi.datamodel.via import ViaProject, RegionFilterConfig, ViaMetadata
from camfi.datamodel.weather import LocationWeatherStationCollector
from camfi.models import model_urls
from camfi.util import Timezone, endpoint_methods, Field
from camfi.transform import RandomHorizontalFlip
from camfi.wingbeat import WingbeatExtractorConfig, extract_all_wingbeats


class ParameterUnspecifiedError(Exception):
    """Base exception called when a parameter which needs to be specified is not."""


class ViaProjectUnspecifiedError(ParameterUnspecifiedError):
    """Raised by CamfiConfig.load_via_project."""


class PlaceUnspecifiedError(ParameterUnspecifiedError):
    """Raised when CamfiConfig.place is requested, but it is unspecified."""


class WingbeatExtractorConfigUnspecifiedError(ParameterUnspecifiedError):
    """Raised when a WingbeatExtractorConfig is needed, but was not supplied."""


class CameraConfigUnspecifiedError(ParameterUnspecifiedError):
    """Raised when a method requies a camera config but none was supplied."""


class AnnotatorConfigUnspecifiedError(ParameterUnspecifiedError):
    """Raised if AnnotatorConfig is required but is not specified."""


class TrainingConfigUnspecifiedError(ParameterUnspecifiedError):
    """Raised if TrainingConfig is required but is not specified."""


class MaskMakerUnspecifiedError(ParameterUnspecifiedError):
    """Raised if MaskMaker is required but is not specified."""


class InferenceConfigUnspecifiedError(ParameterUnspecifiedError):
    """Raised if InferenceConfig is required but is not specified."""


class ValidationConfigUnspecifiedError(ParameterUnspecifiedError):
    """Raised if ValidationConfig is required but is not specified."""


class OutputUnspecifiedError(ParameterUnspecifiedError):
    """Raised if output is required but is not specified."""


class FiltersUnspecifiedError(ParameterUnspecifiedError):
    """Raised if filters is required but is not specified."""


class CameraConfig(BaseModel):
    """Camera hardware-related configuration."""

    camera_time_to_actual_time_ratio: Optional[float] = Field(
        None,
        description=(
            "Used for correcting timestamps from an inaccurate clock. "
            "A value of 1.0 means the camera's clock runs at the correct speed. "
            "A value less than 1.0 means that the camera's clock is slow, "
            "and a value greater than 1.0 means the camera's clock is fast. "
        ),
    )
    line_rate: Optional[PositiveFloat] = Field(
        None,
        description=(
            "Rolling-shutter line rate of camera (lines per second). "
            "Used during wingbeat extraction to determine "
            "the effective exposure time for moving objects "
            "(flying insects) "
            "in images. "
            "This is required in order to calculate wingbeat frequency accurately. "
        ),
    )


class TrainingConfig(BaseModel):
    """Contains settings for camfi annotator model training.

    Parameters
    ----------
    mask_maker : Optional[MaskMaker]
        Instance of MaskMaker which produces instance segementation masks for model
        training.
    min_annotations: int
        Only train on images which have at least this many annotations.
    max_annotations : float
        Only train on images which have at most this many annotations. Can be useful
        if running into memory errors on the GPU, since memory consumption depends on
        the number of annotated objects in the image.
    box_margin : PositiveInt
        Margin to add to bounding boxes of object annotations, for model training.
    test_set_file : Optional[FilePath]
        Alternative to setting test_set directly, load it from a file.
    test_set : list[Path]
        list of images to exclude from training.
    load_pretrained_model : Optional[Union[Path, str]]
        Path or url to model parameters file. If set, will load the pretrained
        parameters. By default, will start with a model pre-trained on the Microsoft
        COCO dataset.
    device : str
        E.g. "cpu" or "cuda". Training is typically much faster on a GPU. Use "cuda" for
        Nvidia GPUs.
    batch_size : int
        Number of images to load at once.
    num_workers : int
        Number of worker processes for data loader to spawn.
    num_epochs : int
        Number of epochs to train.
    outdir : DirectoryPath
        Path to directory where to save model(s).
    model_name : Optional[str]
        Identifier to include in model save file. By default the current date in
        YYYYmmdd format.
    save_intermediate : bool
        If True, model is saved after each epoch, not just after all epoch are complete.
        This is recommended, especially if training on a service which could terminate
        unpredicatbly.
    """

    mask_maker: Optional[MaskMaker] = None
    min_annotations: Optional[int] = Field(
        None,
        description=(
            "Only train on images which have at least this many annotations. "
            "It often makes sense to set this to a low number (e.g. 1). "
            "This will mean that any image with *no* annotations in it "
            "will be skipped. "
        ),
    )
    max_annotations: Optional[int] = Field(
        None,
        description=(
            "Only train on images which have at most this many annotations. "
            "This option exists because GPU memory consumption during "
            "each step of training "
            "depends on the number of annotations in the loaded images, "
            "so images with a lot of annotations can cause training to crash. "
            "It probably makes sense to leave this unset, and only set it if you "
            "are runnning into GPU memory errors. "
        ),
    )
    box_margin: PositiveInt = Field(
        10,
        description=(
            "Margin (in pixels) to add to bounding boxes of object annotations. "
            "This may affect "
            "endpoint estimation of polyline annotataions during inference. "
        ),
    )
    test_set_file: Optional[FilePath] = Field(
        None,
        description=(
            "Path to file "
            "containing filepaths (one per line). Do not set if test_set is set."
        ),
    )
    test_set: list[Path] = Field(
        [],
        description=(
            "list of images to exclude from training. "
            "Also used by validator to determine "
            "which subsets of images to validate against. "
        ),
    )
    load_pretrained_model: Optional[Union[Path, str]] = Field(
        None,
        description=(
            "Path or url to model .pth file or one of camfi.models.model_urls "
            "to initialise training. "
            "By default, Resnet50 FPN backbone trained on the COCO dataset is used. "
        ),
        examples=list(model_urls),
    )
    device: str = Field(
        "cpu",
        description="Device to run training on. Use cuda for a Nvidia GPU.",
        examples=["cuda", "cpu"],
    )
    batch_size: int = Field(
        5,
        description=(
            "Number of images to load at once. "
            "This has memory consumption implications, "
            "So if you're running in to problems with memory availability on the GPU, "
            "consider lowering this. "
        ),
    )
    num_workers: int = Field(
        2,
        description=(
            "Number of worker processes for data loader to spawn. "
            "Depending on your machine, "
            "setting this really high can have diminishing returns "
            "as system IO will become the rate-limiting step. "
        ),
    )
    num_epochs: int = Field(
        10,
        description=(
            "Number of epochs "
            "(traversals of the entire training set) "
            "to train. "
            "Note that data augmentation (e.g. random horizontal flipping) "
            "is applied only once to each image during each epoch. "
        ),
    )
    outdir: DirectoryPath = Field(
        Path(),
        description=(
            "Path to directory where to save model "
            "(or models if save_intermediate is set). "
        ),
    )
    model_name: Optional[str] = Field(
        None,
        description=(
            "Identifier to include in the file name of the trained model(s). "
            "By default, this will be inferred from the date the configuration is "
            "parsed, in the format YYYYmmdd."
        ),
    )
    save_intermediate: bool = Field(
        False,
        description=(
            "If True, model is saved after each epoch. "
            "By default, the model is only saved at the very end. "
        ),
    )

    class Config:
        schema_extra = {
            "description": "Contains settings for camfi annotator model training."
        }

    @validator("test_set", always=True)
    def load_test_set(cls, v, values):
        test_set_file = values.get("test_set_file", None)
        if test_set_file is not None:
            if len(v) > 0:
                raise ValueError("test_set should not be set if test_set_file is set.")
            with open(test_set_file, "r") as f:
                v = [Path(l.strip()) for l in f]
        return v


class InferenceConfig(BaseModel):
    """Contains settings for camfi annotator inference.

    Parameters
    ----------
    output_path : Optional[Path]
        If set, automatically generated annotations will be saved to file.
    model : Union[Path, str]
        Path to .pth file specifying model parameters, model name defined in
        camfi.models.model_urls, or url to model to download from the internet.
    device : str
        Specifies device to run inference on. E.g. set to "cuda" to use an Nvidia GPU.
    backup_device : Optional[str]
        Specifies device to run inference on when a runtime error occurs while using
        device. Probably only makes sense to set this to "cpu" if device="cuda". This
        option enables the annotator to leverage a GPU with limited memory capacity
        without crashing if a difficult image is encountered.
    split_angle : PositiveFloat
        Approximate maximum angle between polyline segments in degrees. Note that this
        will immediately be converted to radians upon instantiation of Annotator.
    poly_order : PositiveInt
        Order of polynomial used for fitting motion blur paths.
    endpoint_method : str
        Method to find endpoints of motion blurs. One of camfi.utils.endpoint_methods.
    endpoint_extra_args : list[Any]
        Extra arguments to pass to endpoint method function.
    score_thresh : float
        Score threshold between 0.0 and 1.0 for automatic annotations to be kept.
    overlap_thresh : float
        Minimum proportion of overlap (weighted intersection over minimum) between two
        instance segmentation masks to infer that one of the masks should be discarded.
    edge_thresh : NonNegativeInt
        Minimum distance an annotation has to be from the edge of the image before it is
        converted from a polyline annotation to a circle annotation.
    """

    output_path: Optional[Path] = Field(
        None,
        description=(
            "If set, automatically generated annotations will be written to "
            "the file at the specified path. "
            "If unset, then the annotations will be written to "
            "the file at the path specified by CamfiConfig.default_output. "
            "If neither is set, "
            "the annotations will not be written to file, "
            "which is probably only useful if you are using "
            "camfi as a python module, "
            "rather than from the command line. "
        ),
    )
    model: Union[Path, str] = Field(
        "release",
        description=(
            "Pre-trained instance segmentation model "
            "to use during automated annotaton. "
            "Can be a "
            "path or url to a model .pth file "
            "or one of camfi.models.model_urls. "
        ),
        examples=list(model_urls),
    )
    device: str = Field(
        "cpu",
        description="Specifies device to run inference on.",
        examples=["cpu", "cuda"],
    )
    backup_device: Optional[str] = Field(
        None,
        description=(
            "Used for images which fail on device due to memory constraints. "
            "It is recommended to set this to 'cpu' if device is set to 'cuda'. "
        ),
    )
    split_angle: PositiveFloat = Field(
        15.0,
        description="Approximate maximum angle between polyline segments in degrees.",
    )
    poly_order: PositiveInt = Field(
        2, description="Order of polynomial used for fitting motion blur paths."
    )
    endpoint_method: str = Field(
        "truncate",
        description="Method to find endpoints of motion blurs.",
        examples=list(endpoint_methods),
    )
    endpoint_extra_args: list[Any] = Field(
        [10], description="Extra arguments to pass to endpoint method function."
    )
    score_thresh: float = Field(
        0.0,
        description=(
            "Score threshold between 0.0 and 1.0 for automatic annotations to be kept."
            "Setting this to higher values will result fewer annotations being made. "
        ),
    )
    overlap_thresh: float = Field(
        0.4,
        description=(
            "Minimum weighted IoM between 0.0 and 1.0 "
            "for non-maximum suppression of detections. "
            "Setting this to higher values "
            "will result in a higher number of annotations being made, "
            "since fewer pairs of annotations will be assumed to relate "
            "to the same object as eachother. "
        ),
    )
    edge_thresh: NonNegativeInt = Field(
        20,
        description=(
            "Polyline annotations which go this close "
            "(measured in pixels) "
            "to edge of the image "
            "will be converted to circle annotations. "
            "This is because wingbeat measurements "
            "cannot be made unless the entire motion blur "
            "is contained within the image. "
        ),
    )

    class Config:
        schema_extra = {
            "description": "Contains settings for camfi annotator inference."
        }

    @validator("endpoint_method")
    def check_endpoint_method(cls, v):
        assert (
            v in endpoint_methods
        ), "endpoint_method must be one of {list(endpoint_methods)}. Got {v}."


class ValidationConfig(BaseModel):
    """Contains settings for camfi annotator validation.

    Parameters
    ----------
    autoannotated_via_project_file : Optional[Path]
        Path to file containing VIA Project with annotations to validate.
    iou_thresh : float
        Threshold of intersection-over-union to match annotations.
    """

    autoannotated_via_project_file: Optional[Path] = Field(
        None,
        description=(
            "Path to file containing VIA Project with "
            "automatically obtained annotations to validate. "
        ),
    )
    iou_thresh: float = Field(
        0.5,
        description=(
            "Threshold of bounding-box intersection-over-union to match "
            "automatically obtained annotations to ground truth annotations. "
        ),
    )
    image_sets: list[str] = Field(
        ["all"],
        description=(
            "Image sets to perform validation on. "
            "Possible sets are 'all', 'test', and 'train'. "
            "Multiple sets can be specified. "
            "If including 'test' or 'train', "
            "annotator.training must also be configured. "
        ),
        regex="^(all|test|train)$",
    )
    output_dir: Optional[DirectoryPath] = Field(
        None,
        description=(
            "If set, results of validation will be saved to this directory "
            "(one file per image set). "
        ),
    )
    output_stem: str = Field(
        "validation",
        description=(
            "Stem of output files. "
            "Output files will be saved as ``output_dir/output_stem.image_set.json``. "
        ),
    )

    class Config:
        schema_extra = {
            "description": "Contains settings for camfi annotator validation."
        }


class AnnotatorConfig(BaseModel):
    """Contains settings for automatic annotation training and inference.

    Parameters
    ----------
    crop : Optional[BoundingBox]
        If specified, images will be cropped to this bounding box after loading them
        from file. This can be useful when working with images taken by trail cameras,
        as some models burn a information bar onto the bottom of the image, which
        ideally should be removed before training or inference on that image.
    training : Optional[TrainingConfig]
        Contains settings for camfi annotator model training.
    inference : Optional[InferenceConfig]
        Contains settings for camfi annotator inference.
    """

    crop: Optional[BoundingBox] = None
    training: Optional[TrainingConfig] = None
    inference: Optional[InferenceConfig] = None
    validation: Optional[ValidationConfig] = None

    class Config:
        schema_extra = {
            "description": "Settings for automatic annotation training and inference."
        }


class ImageFilterConfig(BaseModel):
    """Contains options for filtering images from a VIA project."""

    min_annotations: Optional[int] = Field(
        None,
        description=("Exclude images if they have " "fewer annotations than this. "),
    )
    exclude_images: Optional[list[Path]] = Field(
        None,
        description=(
            "Images to exclude from VIA project. "
            "Can either be a list of paths to image files, "
            "or a single path to a text file (i.e. not a list). "
            "If the latter, image file paths will be read from text file, "
            "one per line."
        ),
    )
    include_images: Optional[list[Path]] = Field(
        None,
        description=(
            "Images to include from VIA project. "
            "Images not in this list will be excluded. "
            "Can either be a list of paths to image files, "
            "or a single path to a text file (i.e. not a list). "
            "If the latter, image file paths will be read from text file, "
            "one per line."
        ),
    )

    @validator("exclude_images", "include_images", pre=True)
    def read_from_file(cls, v):
        if isinstance(v, Path):
            with open(v, "r") as f:
                image_paths = [Path(line.strip()) for line in f]
            return image_paths
        return v


class FilterConfig(BaseModel):
    """Contains settings for filtering images and/or regions (annotations) from a VIA
    project."""

    image_filters: Optional[ImageFilterConfig] = None
    region_filters: Optional[RegionFilterConfig] = None


class CamfiConfig(BaseModel):
    """Defines structure of Camfi's config.json files, and provides methods for loading
    data from various sources.
    """

    root: DirectoryPath = Field(
        Path(), description="Directory containing all images for the project."
    )
    disable_progress_bar: Optional[bool] = Field(
        None,
        description=(
            "Disables progress bars. "
            "By default, disable on non-TTY. "
            "To force the progress bar to appear, "
            "set this to false. "
        ),
    )
    via_project_file: Optional[FilePath] = Field(
        None, description="Path to file containing VIA project."
    )
    day_zero: Optional[date] = Field(
        None,
        description=(
            "Used as reference date for plotting etc. "
            "Currently the Camfi CLI does not access this value, "
            "However it is used in some of the example notebooks. "
        ),
    )
    output_tz: Timezone = Field(
        description=(
            "Sets a global timezone "
            "to convert all timezones to, "
            "for simpler comparison between locations in different timezones. "
        ),
        default_factory=lambda: Timezone("Z"),
    )
    default_output: Optional[Path] = Field(
        None,
        description=(
            "Path to write output. Outputs defined in "
            "wingbeat_extractor and annotator "
            "take precedence over this."
        ),
    )
    filters: Optional[FilterConfig] = None
    camera: Optional[CameraConfig] = None
    time: Optional[LocationTimeCollector] = None
    place: Optional[LocationWeatherStationCollector] = None
    wingbeat_extraction: Optional[WingbeatExtractorConfig] = None
    annotator: Optional[AnnotatorConfig] = None

    @property
    def timestamp_zero(self) -> Optional[pd.Timestamp]:
        if self.day_zero is None:
            return None
        return pd.to_datetime(self.day_zero).tz_localize(self.output_tz._timezone)

    @cached_property
    def via_project(self) -> ViaProject:
        """Loads ViaProject from file. Raises ViaProjectUnspecifiedError if
        self.via_project is None.

        Returns
        -------
        via_project : ViaProject
            VIA project loaded from self.via_project file.
        """
        if self.via_project_file is None:
            raise ViaProjectUnspecifiedError
        return ViaProject.parse_file(self.via_project_file)

    class Config:
        json_encoders = {Timezone: str}
        keep_untouched = (cached_property,)
        schema_extra = {"description": "Camfi configuration."}

    @root_validator
    def check_all_loctions_defined(cls, values):
        if values.get("time") is None:
            return values
        default_place = LocationWeatherStationCollector(
            locations=[], weather_stations=[], location_weather_station_mapping={}
        )
        specified_locations = set(
            location.name for location in values.get("place", default_place).locations
        )
        for camera_placement in values["time"].camera_placements.values():
            if camera_placement.location not in specified_locations:
                raise ValidationError(
                    f"location {camera_placement.location} unspecified in place field."
                )

        return values

    @classmethod
    def parse_yaml(cls, document: str) -> CamfiConfig:
        """Parses yaml document and returns a CamfiConfig instance.

        Parameters
        ----------
        document : str
            StrictYAML document string.

        Returns
        -------
        config : CamfiConfig
            CamfiConfig instance with settings defined in document.
        """
        return cls.parse_obj(load(document).data)

    @classmethod
    def parse_yaml_file(cls, document_path: Path) -> CamfiConfig:
        """Parses yaml document read from file and returns a CamfiConfig instance.

        Parameters
        ----------
        document_path : Path
            Path to file containing yaml document.

        Returns
        -------
        config : CamfiConfig
            CamfiConfig instance with settings defined in document.
        """
        with open(document_path, "r") as f:
            document = f.read()

        return cls.parse_yaml(document)

    def yaml(self, **kwargs) -> str:
        """Serialises self to yaml string.

        Parameters
        ----------
        **kwargs
            Passed to self.json(). E.g. exclude_unset. Note that by default,
            exclude_none=True to avoid YAMLSerializationError. Also, exclude_unset=True
            by default, as this makes more sense for CamfiConfig. This differs from
            the defaults of CamfiConfig.json(). Other defaults are the same.

        Returns
        -------
        document : str
            String containing yaml document.
        """
        kwargs = kwargs.copy()
        kwargs.setdefault("exclude_unset", True)
        if not kwargs.setdefault("exclude_none", True):
            raise NotImplementedError(
                "exclude_none=False is not implemented for yaml. "
                "Consider using json, or leave exclude_none=True."
            )
        config_dict = json.loads(self.json(**kwargs))
        if len(config_dict) == 0:
            return ""
        return as_document(config_dict).as_yaml()

    def apply_image_filters(self) -> None:
        """Applies image filter defined in self.filters to self.via_project.
        Operates in-place.
        """
        if self.filters is None:
            raise FiltersUnspecifiedError

        if self.filters.image_filters is None:
            # Nothing to do.
            return None

        # Define inclusion filter
        if self.filters.image_filters.include_images is None:

            def _passes_include(metadata: ViaMetadata) -> bool:
                return True

        else:
            _include_images = set(self.filters.image_filters.include_images)

            def _passes_include(metadata: ViaMetadata) -> bool:
                return metadata.filename in _include_images

        # Define inclusion filter
        if self.filters.image_filters.exclude_images is None:

            def _passes_exclude(metadata: ViaMetadata) -> bool:
                return True

        else:
            _exclude_images = set(self.filters.image_filters.exclude_images)

            def _passes_exclude(metadata: ViaMetadata) -> bool:
                return metadata.filename not in _exclude_images

        # Define min_annotations filter
        if self.filters.image_filters.min_annotations is None:

            def _passes_min_annotations(metadata: ViaMetadata) -> bool:
                return True

        else:
            _min_annotations = self.filters.image_filters.min_annotations

            def _passes_min_annotations(metadata: ViaMetadata) -> bool:
                return len(metadata.regions) >= _min_annotations

        # Define combined filter
        def _image_filter(metadata: ViaMetadata) -> bool:
            return (
                _passes_include(metadata)
                and _passes_exclude(metadata)
                and _passes_min_annotations(metadata)
            )

        # Apply filter
        self.via_project.filter_inplace(_image_filter)

    def apply_region_filters(self) -> None:
        """Applies region filters defined in self.filters to self.via_project.
        Operates in-place.
        """
        if self.filters is None:
            raise FiltersUnspecifiedError

        if self.filters.region_filters is None:
            # Nothing to do.
            return None

        for metadata in self.via_project.via_img_metadata.values():
            metadata.filter_regions(self.filters.region_filters)

    def get_image_dataframe(self) -> pd.DataFrame:
        """Calls self.via_project.to_image_dataframe(tz=self.output_tz), returning the
        result."""
        if self.place is None:
            raise PlaceUnspecifiedError
        return self.via_project.to_image_dataframe(tz=self.output_tz._timezone)

    def get_weather_dataframe(self) -> pd.DataFrame:
        """Calls self.place.get_weather_dataframe(), returning the result."""
        if self.place is None:
            raise PlaceUnspecifiedError
        return self.place.get_weather_dataframe()

    def get_sun_time_dataframe(
        self, days: Union[str, dict[str, Sequence[date]]]
    ) -> pd.DataFrame:
        """Calls self.place.get_sun_time_dataframe"""
        if self.place is None:
            raise PlaceUnspecifiedError
        if days == "images":
            image_df = self.get_image_dataframe()
            days = {}
            for location in self.place.locations:
                days[location.name] = image_df[image_df["location"] == location.name][
                    "datetime_corrected"
                ].dt.date.unique()

        elif days == "weather":
            weather_df = self.get_weather_dataframe()
            days = {}
            for location in self.place.locations:
                weather_station: str = self.place.location_weather_station_mapping[
                    location.name
                ]
                days[location.name] = weather_df.loc[weather_station].index.date

        if isinstance(days, dict):
            return self.place.get_sun_time_dataframe(days)

        raise TypeError(
            "days must be one of ['images' | 'weather' | dict[str, Sequence[date]]]. "
            f"Got {days} of type {type(days)}."
        )

    def get_merged_dataframe(self) -> pd.DataFrame:
        """"""
        if self.place is None:
            raise PlaceUnspecifiedError
        image_df = self.get_image_dataframe()
        image_df["date"] = pd.to_datetime(image_df["datetime_corrected"].dt.date)

        days = {}
        for location in self.place.locations:
            days[location.name] = image_df[image_df["location"] == location.name][
                "date"
            ].dt.date.unique()

        weather_sun_df = self.place.get_weather_sun_dataframe(days=days)

        merged_df = pd.merge(
            image_df, weather_sun_df, how="left", on=["location", "date"], sort=True
        )
        merged_df.set_index(["location", "date"], inplace=True)

        return merged_df

    def load_all_exif_metadata(self) -> None:
        """Calls self.via_project.load_all_exif_metadata with appropriate arguments,
        set by config. Operates in place.
        """
        time_ratio = None
        if self.camera is not None:
            time_ratio = self.camera.camera_time_to_actual_time_ratio

        location_dict = None
        datetime_correctors = None
        if self.time is not None:
            location_dict = self.time.get_location_dict()
            datetime_correctors = self.time.get_correctors(
                camera_time_to_actual_time_ratio=time_ratio
            )

        self.via_project.load_all_exif_metadata(
            root=self.root,
            location_dict=location_dict,
            datetime_correctors=datetime_correctors,
            disable_progress_bar=self.disable_progress_bar,
        )

    def extract_all_wingbeats(self) -> None:
        """Calls extract_all_wingbeats on self.via_project with parameters taken from
        configuration."""
        if self.wingbeat_extraction is None:
            raise WingbeatExtractorConfigUnspecifiedError
        if self.camera is None or self.camera.line_rate is None:
            raise CameraConfigUnspecifiedError
        extract_all_wingbeats(
            self.via_project,
            root=self.root,
            line_rate=self.camera.line_rate,
            disable_progress_bar=self.disable_progress_bar,
            **self.wingbeat_extraction.dict(),
        )

    @property
    def training_dataset(self) -> CamfiDataset:
        """Gets CamfiDataset suitable for training camfi automatic annotation model."""
        if self.annotator is None:
            raise AnnotatorConfigUnspecifiedError

        if self.annotator.training is None:
            raise TrainingConfigUnspecifiedError

        if self.annotator.training.mask_maker is not None:
            mask_maker = self.annotator.training.mask_maker
        elif self.annotator.crop is not None:
            mask_maker = MaskMaker(shape=self.annotator.crop.shape)
        else:
            raise MaskMakerUnspecifiedError(
                "Need to set annotator.training.mask_maker and/or annotator.crop"
            )

        return CamfiDataset(
            root=self.root,
            via_project=self.via_project,
            crop=self.annotator.crop,
            inference_mode=False,
            mask_maker=mask_maker,
            transform=RandomHorizontalFlip(prob=0.5),
            min_annotations=self.annotator.training.min_annotations,
            max_annotations=self.annotator.training.max_annotations,
            box_margin=self.annotator.training.box_margin,
            exclude=set(self.annotator.training.test_set),
        )

    def train_model(self) -> Path:
        """Calls camfi.annotator.train_model with appropriate arguments from self."""
        if self.annotator is None:
            raise AnnotatorConfigUnspecifiedError

        if self.annotator.training is None:
            raise TrainingConfigUnspecifiedError

        return train_model(
            self.training_dataset,
            load_pretrained_model=self.annotator.training.load_pretrained_model,
            device=self.annotator.training.device,
            batch_size=self.annotator.training.batch_size,
            num_workers=self.annotator.training.num_workers,
            num_epochs=self.annotator.training.num_epochs,
            outdir=self.annotator.training.outdir,
            model_name=self.annotator.training.model_name,
            save_intermediate=self.annotator.training.save_intermediate,
        )

    @property
    def inference_dataset(self) -> CamfiDataset:
        """Gets CamfiDataset suitable for automatic annotaton."""
        if self.annotator is None:
            raise AnnotatorConfigUnspecifiedError

        return CamfiDataset(
            root=self.root,
            via_project=self.via_project,
            crop=self.annotator.crop,
            inference_mode=True,
        )

    def annotate(self) -> ViaProject:
        """Performs automatic annotation on all the images in project. Saves project to
        a file if self.annotator.inference.output_path is set.

        Returns
        -------
        project : ViaProject
            With automatic annotations made.
        """
        if self.annotator is None:
            raise AnnotatorConfigUnspecifiedError

        if self.annotator.inference is None:
            raise InferenceConfigUnspecifiedError

        annotator = Annotator(
            dataset=self.inference_dataset,
            model=self.annotator.inference.model,
            device=self.annotator.inference.device,
            backup_device=self.annotator.inference.backup_device,
            split_angle=self.annotator.inference.split_angle,
            poly_order=self.annotator.inference.poly_order,
            endpoint_method=endpoint_methods[self.annotator.inference.endpoint_method],
            endpoint_extra_args=self.annotator.inference.endpoint_extra_args,
            score_thresh=self.annotator.inference.score_thresh,
            overlap_thresh=self.annotator.inference.overlap_thresh,
            edge_thresh=self.annotator.inference.edge_thresh,
        )
        annotated_project = annotator.annotate(
            disable_progress_bar=self.disable_progress_bar
        )

        # Optionally save ouptut to file
        output_path = (
            self.annotator.inference.output_path
            if self.annotator.inference.output_path
            else self.default_output
        )
        if output_path:
            with open(output_path, "w") as f:
                f.write(annotated_project.json(indent=2))

        return annotated_project

    def get_autoannotated_via_project(self):
        """Loads automatic annotations from file.

        Returns
        -------
        project : ViaProject
            ViaProject with automatically aquired annotations.
        """
        if self.annotator is None:
            raise AnnotatorConfigUnspecifiedError

        if (
            self.annotator.validation is not None
            and self.annotator.validation.autoannotated_via_project_file is not None
        ):
            return ViaProject.parse_file(
                self.annotator.validation.autoannotated_via_project_file
            )

        if (
            self.annotator.inference is None
            or self.annotator.inference.output_path is None
        ):
            raise ValidationConfigUnspecifiedError(
                "Could not determine which file to validate."
            )

        return ViaProject.parse_file(self.annotator.inference.output_path)

    def validate_annotations(self) -> list[AnnotationValidationResult]:
        """Validates automatically aquired annotations against ground-truth annotations.

        Returns
        -------
        validation_results : list[AnnotationValidationResult]
            Results from validation.
        """
        if self.annotator is None:
            raise AnnotatorConfigUnspecifiedError
        annotator: AnnotatorConfig = self.annotator

        if annotator.validation is None:
            raise ValidationConfigUnspecifiedError

        training: TrainingConfig

        subset_functions = {}
        for image_set in annotator.validation.image_sets:
            if image_set == "all":
                subset_functions["all"] = lambda x: True
            elif image_set == "train":
                if annotator.training is None:
                    raise TrainingConfigUnspecifiedError
                else:
                    training = annotator.training
                    exclude_set = set(training.test_set)

                    def _subset_fn(metadata: ViaMetadata) -> bool:
                        if metadata.filename in exclude_set:
                            return False
                        if (
                            training.min_annotations is not None
                            and len(metadata.regions) < training.min_annotations
                        ):
                            return False
                        if (
                            training.max_annotations is not None
                            and len(metadata.regions) > training.max_annotations
                        ):
                            return False
                        return True

                    subset_functions["train"] = _subset_fn
            elif image_set == "test":
                if annotator.training is None:
                    raise TrainingConfigUnspecifiedError
                else:
                    include_set = set(training.test_set)
                    training = annotator.training

                    def _subset_fn(metadata: ViaMetadata) -> bool:
                        if metadata.filename not in include_set:
                            return False
                        if (
                            training.min_annotations is not None
                            and len(metadata.regions) < training.min_annotations
                        ):
                            return False
                        if (
                            training.max_annotations is not None
                            and len(metadata.regions) > training.max_annotations
                        ):
                            return False
                        return True

                    subset_functions["test"] = _subset_fn
            else:
                raise ValueError(f"Expected one of all|train|test. Got {image_set}.")

        validation_results = validate_annotations(
            auto_annotations=self.get_autoannotated_via_project(),
            ground_truth=self.via_project,
            iou_thresh=annotator.validation.iou_thresh,
            subset_functions=subset_functions,
            disable_progress_bar=self.disable_progress_bar,
        )

        # Optionally save to file before returning
        output_dir = (
            annotator.validation.output_dir
            if annotator.validation.output_dir
            else self.default_output
        )
        if output_dir:
            for image_set, validation_result in zip(
                annotator.validation.image_sets, validation_results
            ):
                output_file = (
                    output_dir / f"{annotator.validation.output_stem}.{image_set}.json"
                )
                with open(output_file, "w") as f:
                    f.write(validation_result.json(indent=2))

        return validation_results

    def write_project(self) -> None:
        """Writes self.via_project as json to self.default_output (if set) or stdout."""
        if self.default_output:
            with open(self.default_output, "w") as f:
                f.write(self.via_project.json(indent=2, exclude_unset=True))

        else:
            print(self.via_project.json(indent=2, exclude_unset=True))

    def filelist(self) -> list[Path]:
        """lists the images in self.via_project.

        Returns
        -------
        image_files : list[Path]
            list of filepaths.
        """
        return [m.filename for m in self.via_project.via_img_metadata.values()]

    def zip_images(self) -> None:
        """Makes a zip archive of all the images in the VIA project file.
        Requires self.default_output to be set.
        """
        if self.default_output is None:
            raise OutputUnspecifiedError

        image_files = self.filelist()

        with ZipFile(self.default_output, mode="w") as f:
            for filename in tqdm(
                image_files,
                disable=self.disable_progress_bar,
                desc="Zipping images",
                unit="img",
                dynamic_ncols=True,
                ascii=True,
                color="green",
            ):
                f.write(filename)



"""Provides ``model_urls``, a ``dict`` mapping model name strings to URLs which point to
pre-trained Camfi autoannotation models.

Currently defined models:

v1
    Trained on the `2019 Cabramurra dataset <https://doi.org/10.5281/zenodo.4950570>`_
    using Camfi v1.2. Should still be compatible with Camfi v2, but this is untested.

release
    Can change with each release of Camfi.
    Currently set to **v1**.
"""
_release_model = "v1"
model_urls = {
    "v1": "https://github.com/J-Wall/camfi/releases/download/1.0/20210519_4_model.pth",
}

model_urls["release"] = model_urls[_release_model]



"""This module can be ignored unless you are developing camfi.

The code in this file is not directly used by camfi. It is used for generating
_region_filter_config_static.py which is imported by via.py.

Running ``python camf/datamodel/_region_filter_config_dynamic.py`` from the camfi
root dir will generate region_filter_config.py. The funky thing about this
is this script actually depends on camfi being installed already. Make of that what you
will. Basically, this script can be ignored unless you are developing the
``ViaRegionAttributes`` class specifically, in which case you should run this script
after you make any changes.

region_filter_config.py should not be edited manually.
"""

from numbers import Real
from pathlib import Path
from typing import Optional

from pydantic import BaseModel, create_model

from camfi.datamodel.via_region_attributes import ViaRegionAttributes
from camfi.util import Field


class FloatFilter(BaseModel):
    ge: float = Field(
        ...,
        title="Greater-than or Equal-to",
        description="Only include region if attribute >= this value.",
    )
    le: float = Field(
        ...,
        title="Less-than or Equal-to",
        description="Only include region if attribute <= this value.",
    )
    exclude_none: bool = Field(
        False, description="Whether to exclude region if attribute is not set."
    )


_region_filter_config_fields = {
    name: (
        Optional[FloatFilter],
        Field(None, description=f"Sets filters for the {name} region attribute."),
    )
    for name in ViaRegionAttributes.__fields__.keys()
    if issubclass(ViaRegionAttributes.__fields__[name].type_, Real)
}

RegionFilterConfig = create_model(  # type: ignore[var-annotated]
    "RegionFilterConfig",
    **_region_filter_config_fields,  # type: ignore[arg-type]
)


if __name__ == "__main__":

    # Deferred imports
    from datamodel_code_generator import InputFileType, generate

    # Generate schema
    json_schema = RegionFilterConfig.schema_json()

    # Generate _region_filter_config_static.py
    output = Path(__file__).parent / "region_filter_config.py"
    generate(
        json_schema,
        input_file_type=InputFileType.JsonSchema,
        input_filename=Path(__file__).name,
        output=output,
    )

    # Add config to RegionFilterConfig
    desc = (
        "Contains options for filtering regions (annotations) from images, "
        "based on the values of region attributes. "
    )
    with open(output, "a") as f:
        f.write(
            f"""
    class Config:
        schema_extra = {{
            "description": {desc!r}
        }}"""
        )



"""Defines command-line interface to Camfi.
"""

from argparse import ArgumentParser, RawDescriptionHelpFormatter as Formatter
from datetime import datetime
from functools import wraps
from inspect import getdoc
from pathlib import Path
from shutil import get_terminal_size
from sys import exit, stderr
from textwrap import fill
from typing import Callable, Optional

from camfi import __version__

CONFIG_URL = "https://camfi.readthedocs.io/en/latest/usage/configuration.html"


def _qprint(*args, **kwargs) -> None:
    pass


def _vprint(*args, **kwargs) -> None:
    print(f"[{datetime.now().astimezone().isoformat()}]", *args, file=stderr, **kwargs)


def _fill_paras(paras: list[str], sep: str = "\n \n", **kwargs) -> str:
    return sep.join(
        para if para.startswith("  ") else fill(para, **kwargs) for para in paras
    )


class ConfigParseError(Exception):
    """Raised if config parsing fails."""


class Commander:
    """Defines commands for camfi cli. Any command defined on this class can be called
    by providing them as command-line arguments to ``camfi``.
    """

    _to_sh = {ord("_"): ord("-")}
    _to_py = {ord("-"): ord("_")}

    def __init__(
        self,
        config_path: Optional[Path],
        input_file: Optional[Path] = None,
        output: Optional[Path] = None,
        disable_progress_bar: Optional[bool] = None,
        vprint: Callable = _vprint,
        vvprint: Callable = _qprint,
    ):
        """Parses config_path file into self.config : camfi.projectconfig.CamfiConfig.

        Parameters
        ----------
        config_path : Optional[Path]
            Path to configuration file. Can be JSON (.json) or StrictYAML (.yaml|.yml).
        output : Optional[Path]
            If set, ``self.default_output`` will be overwritten with this value.
        """
        self._vprint, self._vvprint = vprint, vvprint

        # Import only done after command line parsing to make it faster and more robust.
        self._vvprint("Importing camfi...")
        from camfi.projectconfig import CamfiConfig

        self._vvprint("Done.")

        self._vprint(
            f"Parsing configuration file: {config_path}..."
            if config_path
            else "Creating configuration."
        )
        if config_path is None:
            self.config = CamfiConfig()
        elif config_path.suffix.startswith(".j"):
            self.config = CamfiConfig.parse_file(config_path)
        elif config_path.suffix.startswith(".y"):
            self.config = CamfiConfig.parse_yaml_file(config_path)
        else:
            raise ConfigParseError(
                "Could not determine config format from file suffix. "
                f"Expected one of (.json|.yaml|.yml). Got {config_path.suffix}."
            )
        self._vprint("Done.")

        self._vvprint("Updating command-line configurable config params...")
        if input_file:
            self._vvprint(f"Setting config.via_project_file = {input_file}")
            self.config.via_project_file = input_file
        if output:
            self._vvprint(f"Setting config.default_output = {output}")
            self.config.default_output = output
        if disable_progress_bar is not None:  # Could be True or False
            self._vvprint(
                f"Setting config.disable_progress_bar = {disable_progress_bar}"
            )
            self.config.disable_progress_bar = disable_progress_bar
        self._vvprint("Done updating params.")

    @classmethod
    def _get_command(cls, command: str) -> Callable[[], None]:
        return getattr(cls, command.translate(cls._to_py))

    def __call__(self, command: str) -> None:
        """Calls method of self corresponding to command.

        Parameters
        ----------
        command : str
            Name of command to call. These are the same as the methods defined on this
            class. Any "-" characters are converted to "_" charaters.
        """
        return self._get_command(command)(self)  # type: ignore[call-arg]

    def _write(self, s: str):
        if self.config.default_output:
            with open(self.config.default_output, "w") as f:
                f.write(s)
        else:
            print(s)

    @classmethod
    def cmds(cls) -> dict[str, Optional[str]]:
        """Returns the commands defined by this class.

        Returns
        -------
        commands : dict[str, str]
            Dictionary with command names mapped to command docstrings.
        """
        return {
            choice.translate(cls._to_sh): getdoc(cls._get_command(choice))
            for choice in filter(
                lambda x: not (x.startswith("_") or x == "cmds"), dir(cls)
            )
        }

    def annotate(self) -> None:
        """Performs automatic annotation on all the images in via_project, outputting
        the resulting annotated VIA project file to the configured ``output_path``
        specified under ``annotator.inference`` in the configuration file.
        Requires ``annotator.inference`` to be configured.
        While not strictly required, ``annotator.inference.output_path`` should be
        configured, otherwise the result of annotation will not be saved before the
        program terminates (and this is *probably* not what you want).
        Alternatively, you can configure ``default_output`` either in the configuration
        file or by using the ``-o``/``--output`` flag.
        """
        self.config.annotate()

    def train(self) -> None:
        """Trains a camfi instance segmentation annotation model on manually annotated
        dataset, saving to trained model to the ``outdir`` configured under
        ``annotator.training``. Requires ``annotator.training`` to be configured.
        If ``annotator.inference`` is configured, but under it ``model`` has not been
        explicitely set, then after the model is trained ``model`` of
        ``annotator.inference`` will be set to the newly trained model.
        This means that with proper configuration, ``train`` and ``annotate``
        can be strung together in one command (i.e. ``camfi train annotate``).
        """
        model_path = self.config.train_model()

        # Update model used for annotation
        if self.config.annotator and self.config.annotator.inference:
            if "model" not in self.config.annotator.inference.__fields_set__:
                self._vprint(f"Setting config.annotator.inference.model = {model_path}")
                self.config.annotator.inference.model = model_path
            else:
                self._vprint(
                    "Inference model already set to "
                    f"{self.config.annotator.inference.model} "
                    "Not changing."
                )

    def validate(self) -> None:
        """Validates automatically aquired annotations against ground-truth annotations,
        saving the results to the ``output_dir``
        configured under ``annotator.validation``.
        Requires ``annotator.validation`` to be configured.
        While not strictly required, ``annotator.validation.output_dir``
        should be configured, otherwise the result of validation will not be saved
        before the program terminates (and this is *probably* not what you want).
        If ``image_set`` under ``annotator.validation`` contains
        "train" or "test",
        then ``test_set`` (or ``test_set_file``) under ``annotator.training`` should
        also be configured.
        If ``annotator.training`` is not set,
        then the "train" ``image_set`` will be equivalent to "all",
        and "test" will be an empty set of images.
        It is also possible to leave ``autoannotated_via_project_file`` under
        ``annotator.validation`` unconfigured. In this case, the ``output_path`` from
        ``annotation.inference`` will be validated
        (so at least one of these must be configured).
        Alternatively, you can configure ``default_output`` either in the configuration
        file or by using the ``-o``/``--output`` flag.
        """
        self.config.validate_annotations()

    def load_exif(self) -> None:
        """Loads EXIF metadata into VIA project in-place after reading it from file.
        If ``time`` (and optionally ``camera``) are configured,
        then this will also insert location and corrected timestamp metadata.
        """
        self.config.load_all_exif_metadata()

    def extract_wingbeats(self) -> None:
        """Runs the Camfi algorithm to
        extract wingbeat data from all images in the VIA project,
        inserting that data into the project in-place.
        Requires ``camera`` and ``wingbeat_extraction`` to be configured.
        """
        self.config.extract_all_wingbeats()

    def filter_images(self) -> None:
        """Applies filters to exclude images from VIA project.
        Operates in-place on the VIA project.
        Does nothing if ``filters.image_filters`` isn't configured.
        """
        if self.config.filters is None or self.config.filters.image_filters is None:
            self._vprint("No image filters configured. Skipping.")
            return None
        self.config.apply_image_filters()

    def filter_regions(self) -> None:
        """Applies filters to exclude regions (annotations) from VIA project.
        Operates in-place on the VIA project.
        Does nothing if ``filters.region_filters`` isn't configured.
        """
        if self.config.filters is None or self.config.filters.region_filters is None:
            self._vprint("No region filters configured. Skipping.")
            return None
        self.config.apply_region_filters()

    def write(self) -> None:
        """Writes VIA project to stdout or file
        (set using ``default_output`` configuration parameter or ``-o``/``--output``).
        Can be used after other commands which act in-place on the VIA project
        (e.g. ``load-exif``, ``extract-wingbeats`` and ``apply-filters``).
        Prints to stdout if no output is given.
        """
        self.config.write_project()

    def do_nothing(self) -> None:
        """Does nothing, except parse options and configuration.
        This can be useful if all you want to do is validate and/or convert the
        configuration file.
        This is the default command which is run when ``camfi`` is called.
        """
        pass

    def filelist(self) -> None:
        """Lists the images in the VIA project to stdout or file
        (set using ``default_output`` configuration parameter or ``-o``/``--output``).
        """
        image_files = self.config.filelist()
        self._write("\n".join(str(image_file) for image_file in image_files))

    def region_table(self) -> None:
        """Produces a table with one row per region (annotation).
        Table is written to stout or file
        (set using ``default_output`` configuration parameter or ``-o``/``--output``).
        """
        region_df = self.config.via_project.to_region_dataframe()
        self._write(region_df.to_csv(sep="\t", index=False))

    def image_table(self) -> None:
        """Produces a table with one row per image, with various image metadata columns,
        including n_annotations (the number of annotations in the image).
        Table is written to stout or file
        (set using ``default_output`` configuration parameter or ``-o``/``--output``).
        """
        image_df = self.config.get_image_dataframe()
        self._write(image_df.to_csv(sep="\t", index=False))

    def table(self) -> None:
        """Similar to ``image-table``, but includes weather and sun time columns.
        Produces a table with one row per image, with various image metadata columns,
        including n_annotations (the number of annotations in the image).
        Table is written to stout or file
        (set using ``default_output`` configuration parameter or ``-o``/``--output``).
        """
        image_df = self.config.get_merged_dataframe()
        self._write(image_df.to_csv(sep="\t"))

    def zip_images(self) -> None:
        """Makes a zip archive of all the images in the VIA project file
        (``default_output`` configuration parameter or ``-o``/``--output`` must be set).
        """
        self.config.zip_images()


def _get_description(show_rst: bool, terminal_width: int) -> str:
    description = [
        (
            f"Camfi v{__version__}. "
            "Copyright 2021 Jesse Rudolf Amenuvegbe Wallace "
            "and contributors. "
            "Licenced under the MIT Licence. "
            "Full documentation available at "
            "https://camfi.readthedocs.io/en/latest/. "
            "Source code available from "
            "https://github.com/J-Wall/camfi. "
            "Cite as "
        )
        if not show_rst
        else "",
        (
            "    Wallace (2021). J-Wall/camfi. Zenodo.\n"
            "    https://doi.org/10.5281/zenodo.4971144."
        )
        if not show_rst
        else "",
        (
            "Camfi is a method "
            "for the long-term non-invasive monitoring "
            "of the activity "
            "and abundance "
            "of low-flying insects "
            "using inexpensive wildlife cameras. "
            "It provides utilities "
            "for measuring "
            "the wingbeat frequency "
            "of insects "
            "in still images, "
            "based on the motion blurs "
            "drawn on the image sensor "
            "by the insect "
            "moving through the air. "
            "For large-scale monitoring projects, "
            "camfi enables "
            "automatic annotation "
            "of flying insects "
            "using the Mask R-CNN framework."
        )
        if not show_rst
        else "",
        (
            "Most configuration "
            "for camfi is done with a "
            "configuration file "
            "rather than with "
            "command-line arguments and options. "
            "Documentation for the configuration can be found here: "
            f"{':doc:`configuration`' if show_rst else CONFIG_URL}. "
        ),
        (
            "Camfi makes a bunch of sub-commands available to the command line "
            "(listed below). "
            "If multiple commands are given to a single camfi command, "
            "they will be executed in sequence. "
            "For example, "
            f"the command::"
        ),
        (
            "    $ camfi \\\n"
            "          --input project.json --root datadir \\\n"
            "          --output project_with_wingbeats.json \\\n"
            "          load-exif extract-wingbeats write"
        ),
        (
            "will first load a VIA project from a file, "
            "then load EXIF metadata from the image files (in datadir) "
            "into the project (``load-exif``), "
            "then the camfi wingbeat extraction algoritm will be run to obtain "
            "wingbeat frequencies (``extract-wingbeats``), "
            "and finally the project will be "
            "written to ``project_with_wingbeats.json`` (``write``). "
        ),
        ("Another command::"),
        ("    $ camfi train annotate validate"),
        (
            "would load a VIA project from a file, "
            "train an instance segmentation model "
            "using the images and (manually obtained) annotations "
            "in the project, "
            "and save the new model to disk (``train``). "
            "Then camfi would re-annotate (``annotate``) the images in the project "
            "using the newly trained model, "
            "saving the newly obtained annotations to a new file "
            "(assuming ``annotator`` is properly configured). "
            "Finally, camfi would validate the automatically-obtained annotations "
            "against the manually-obtained ones (``validate``). "
            "All this, while properly handling keeping "
            "training and test image sets separate "
            "where required (and as per the configuration)."
        ),
    ]

    description_str = _fill_paras(description, width=terminal_width)
    return "".join(description_str.split("``"))


def _get_epilog(
    commands: dict[str, Optional[str]], show_rst: bool, terminal_width: int
) -> str:
    epilog = [
        ("Below are the list of commands available to Camfi.\n")
        if show_rst
        else "available commands: "
    ]
    for command, docs in commands.items():
        if not docs:
            docs = "Undocumented."
        if show_rst:
            epilog.append(f"\n{command}\n    {fill(docs, subsequent_indent='    ')}")
        else:
            _command = (
                f"  {command:7}"
                if len(command) < 7
                else f"  {command:{terminal_width - 2}}"
            )
            epilog.append(
                fill(
                    f"{_command} {''.join(docs.split('``'))}",
                    width=terminal_width,
                    subsequent_indent=f"{'':10}",
                )
            )

    return "\n".join(epilog)


def get_argument_parser(show_rst: bool = True) -> ArgumentParser:
    """Defines arguments to the ``camfi`` command.

    Parameters
    ----------
    show_rst : bool
       If False, reStructuredText will be ommitted from description and epilog.

    Returns
    -------
    parser : ArgumentParser
        Command-line argument parser for ``camfi``.
    """
    commands = Commander.cmds()
    terminal_width = get_terminal_size().columns

    parser = ArgumentParser(
        prog="camfi",
        description=_get_description(show_rst, terminal_width),
        epilog=_get_epilog(commands, show_rst, terminal_width),
        formatter_class=Formatter,
    )

    parser.add_argument(
        "--version",
        action="store_true",
        help="Show version and exit.",
    )
    parser.add_argument(
        "-c",
        "--config",
        metavar="path",
        type=Path,
        help=(
            "Path to configuration file. "
            "Can be JSON (.json) or StrictYAML (.yaml|.yml). "
            "If no configuration file is supplied, "
            "a default (empty) configuration is used. "
            "Most commands require at least some configuration. "
        ),
    )
    parser.add_argument(
        "-i",
        "--input",
        metavar="path",
        type=Path,
        help=(
            "Path to input VIA project file. "
            "Replaces ``via_project_file`` in configuration file, "
            "providing an alternative to setting it in the configuration file. "
        ),
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="path",
        type=Path,
        help=(
            "Path to output file. "
            "Replaces ``default_output`` in configuration file, "
            "providing an alternative to setting specific outputs in the configuration "
            "file. "
        ),
    )
    parser.add_argument(
        "-r",
        "--root",
        metavar="dir",
        type=Path,
        help=(
            "Directory containing all images for the project. "
            "Replaces ``root`` in configuration file. "
        ),
    )
    parser.add_argument(
        "-d",
        "--disable-progress-bar",
        action="store_const",
        const=True,
        help="Disables progress bars. By default, disable on non-TTY. ",
    )
    parser.add_argument(
        "-p",
        "--progress-bar",
        action="store_const",
        const=True,
        help="Forces progress bars. By default, disable on non-TTY. ",
    )
    parser.add_argument(
        "-j",
        "--json-conf-out",
        metavar="path",
        type=Path,
        help=(
            "If set, configuration will be written "
            "to file in JSON format after it is parsed. "
            "Set to - to have config written to stdout. "
        ),
    )
    parser.add_argument(
        "-y",
        "--yaml-conf-out",
        metavar="path",
        type=Path,
        help=(
            "If set, configuration will be written "
            "to file in StrictYAML format after it is parsed. "
            "Set to - to have config written to stdout. "
        ),
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="Suppress cli info. You may also like to use ``-d``. ",
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Extra verbose cli info. "
    )
    parser.add_argument(
        "commands",
        nargs="*",
        type=str,
        choices=commands.keys(),
        metavar="command",
        default="do-nothing",
        help=(
            "One or more commands to be executed sequentially by camfi. "
            f"{'' if show_rst else f'Can be one of {set(commands.keys())}. '}"
            "Each command uses the same configuration, "
            "as specified by "
            f"{'``-c``/``--config``' if show_rst else '-c/--config'}. "
        ),
    )

    return parser


def main():
    parser = get_argument_parser(show_rst=False)
    args = parser.parse_args()

    if args.version:
        print(f"Camfi v{__version__}")
        exit(0)

    # Set vprint and vvprint
    if args.verbose:
        vprint, vvprint = _vprint, _vprint
    elif args.quiet:
        vprint, vvprint = _qprint, _qprint
    else:
        vprint, vvprint = _vprint, _qprint

    disable_progress_bar = args.disable_progress_bar
    if args.progress_bar:
        disable_progress_bar = False

    commander = Commander(
        args.config,
        input_file=args.input,
        output=args.output,
        disable_progress_bar=disable_progress_bar,
        vprint=vprint,
        vvprint=vvprint,
    )

    if str(args.json_conf_out) == "-":
        vprint("Writing config JSON to stdout...")
        print(commander.config.json(indent=2, exclude_unset=True))
        vprint("Done.")
    elif args.json_conf_out:
        vprint(f"Writing config JSON to {args.json_conf_out}...")
        with open(args.json_conf_out, "w") as f:
            f.write(commander.config.json(indent=2, exclude_unset=True))
        vprint("Done.")

    # Output config YAML
    if str(args.yaml_conf_out) == "-":
        vprint("Writing config YAML to stdout...")
        print(commander.config.yaml())
        vprint("Done.")
    elif args.yaml_conf_out:
        vprint(f"Writing config YAML to {args.yaml_conf_out}...")
        with open(args.yaml_conf_out, "w") as f:
            f.write(commander.config.yaml())
        vprint("Done.")

    # Run commands
    commands = args.commands if isinstance(args.commands, list) else [args.commands]
    for command in commands:
        vprint(f"Running command: {command}")
        commander(command)
        vprint("Done.")


if __name__ == "__main__":
    main()



# generated by datamodel-codegen:
#   filename:  _region_filter_config_dynamic.py
#   timestamp: 2021-08-07T02:07:17+00:00

from __future__ import annotations

from typing import Optional

from pydantic import BaseModel, Field


class FloatFilter(BaseModel):
    ge: float = Field(
        ...,
        description="Only include region if attribute >= this value.",
        title="Greater-than or Equal-to",
    )
    le: float = Field(
        ...,
        description="Only include region if attribute <= this value.",
        title="Less-than or Equal-to",
    )
    exclude_none: Optional[bool] = Field(
        False,
        description="Whether to exclude region if attribute is not set.",
        title="Exclude None",
    )


class RegionFilterConfig(BaseModel):
    score: Optional[FloatFilter] = Field(
        None, description="Sets filters for the score region attribute.", title="Score"
    )
    best_peak: Optional[FloatFilter] = Field(
        None,
        description="Sets filters for the best_peak region attribute.",
        title="Best Peak",
    )
    blur_length: Optional[FloatFilter] = Field(
        None,
        description="Sets filters for the blur_length region attribute.",
        title="Blur Length",
    )
    snr: Optional[FloatFilter] = Field(
        None, description="Sets filters for the snr region attribute.", title="Snr"
    )
    wb_freq_up: Optional[FloatFilter] = Field(
        None,
        description="Sets filters for the wb_freq_up region attribute.",
        title="Wb Freq Up",
    )
    wb_freq_down: Optional[FloatFilter] = Field(
        None,
        description="Sets filters for the wb_freq_down region\n\nattribute.",
        title="Wb Freq Down",
    )
    et_up: Optional[FloatFilter] = Field(
        None, description="Sets filters for the et_up region attribute.", title="Et Up"
    )
    et_dn: Optional[FloatFilter] = Field(
        None, description="Sets filters for the et_dn region attribute.", title="Et Dn"
    )

    class Config:
        schema_extra = {
            "description": "Contains options for filtering regions (annotations) from images, based on the values of region attributes. "
        }



"""Defines procedures for training, and evaluation automatic camfi annotation models,
and for using them for making automatic annotations (inference). Depends on camfi.util,
camfi.datamodel.autoannotation, camfi.datamodel.geometry, camfi.datamode.via, as well
as ._torchutils and ._models."""

from datetime import datetime
import itertools
from math import pi
from pathlib import Path
from typing import Any, Callable, Optional, Union
from sys import stderr

import numpy as np
from pydantic import (
    BaseModel,
    DirectoryPath,
    NonNegativeInt,
    NonNegativeFloat,
    PositiveFloat,
    PositiveInt,
    validator,
)
from scipy import sparse
import torch
from torch.utils.data import DataLoader
from torchvision.models.detection.mask_rcnn import MaskRCNN
from tqdm import tqdm, trange

from camfi.datamodel.autoannotation import CamfiDataset, Prediction
from camfi.datamodel.geometry import (
    BoundingBox,
    CircleShapeAttributes,
    PolylineShapeAttributes,
)
from camfi.datamodel.via import (
    ViaFileAttributes,
    ViaMetadata,
    ViaProject,
    ViaRegion,
    ViaRegionAttributes,
)
from camfi.models import model_urls
from camfi.util import endpoint_truncate, weighted_intersection_over_minimum, Field
from ._torchutils import collate_fn, get_model_instance_segmentation, train_one_epoch


def load_annotation_model(model_path_or_url: Union[Path, str]) -> MaskRCNN:
    """Loads a camfi annotation model. Accepts any model key provided in
    camfi.models, a Path object, or a URL str.

    Parameters
    ----------
    model_path_or_url : Union[Path, str]
        Path to .pth file specifying model parameters, model name defined in
        camfi.models.model_urls, or url to model to download from the internet.

    Returns
    -------
    model : MaskRCNN
        Instance segmentation model used for automatic annotation.
    """
    print(f"Loading model: {model_path_or_url}", file=stderr)
    model = get_model_instance_segmentation(2, pretrained=False)
    if isinstance(model_path_or_url, Path):
        state_dict = torch.load(model_path_or_url)
    elif model_path_or_url in model_urls:
        state_dict = torch.hub.load_state_dict_from_url(model_urls[model_path_or_url])
    else:
        state_dict = torch.hub.load_state_dict_from_url(model_path_or_url)
    model.load_state_dict(state_dict)
    return model


def copy_annotation_model(model: MaskRCNN) -> MaskRCNN:
    """Copies a camfi annotation model.

    Parameters
    ----------
    model : MaskRCNN
        Model to copy.

    Returns
    -------
    model_copy : MaskRCNN
        Copy of model.
    """
    model_copy = get_model_instance_segmentation(2, pretrained=False)
    model_copy.load_state_dict(model.state_dict())
    return model_copy


def train_model(
    dataset: CamfiDataset,
    load_pretrained_model: Optional[Union[Path, str]] = None,
    device: Union[str, torch.device] = "cpu",
    batch_size: int = 5,
    num_workers: int = 2,
    num_epochs: int = 10,
    outdir: DirectoryPath = Path(),
    model_name: Optional[str] = None,
    save_intermediate: bool = False,
) -> Path:
    """Trains a camfi instance segmentation annotation model on specified dataset,
    saving to trained model to outdir.

    Parameters
    ----------
    dataset : CamfiDataset
        Dataset on which to train the model.
    load_pretrained_model : Optional[Union[Path, str]]
        Path or url to model parameters file. If set, will load the pretrained
        parameters. By default, will start with a model pre-trained on the Microsoft
        COCO dataset.
    device : Union[str, torch.device]
        E.g. "cpu" or "cuda". Training is typically much faster on a GPU. Use "cuda" for
        Nvidia GPUs.
    batch_size : int
        Number of images to load at once.
    num_workers : int
        Number of worker processes for data loader to spawn.
    num_epochs : int
        Number of epochs to train.
    outdir : DirectoryPath
        Path to directory where to save model(s).
    model_name : Optional[str]
        Identifier to include in model save file. By default the current date in
        YYYYmmdd format.
    save_intermediate : bool
        If True, model is saved after each epoch, not just after all epoch are complete.
        This is recommended, especially if training on a service which could terminate
        unpredicatbly (e.g. Google Colab).

    Returns
    -------
    model_path : Path
        Path to saved model.
    """
    # Parameter setting
    device = torch.device(device)
    if model_name is None:
        model_name = datetime.now().strftime("%Y%m%d")

    # Initialise data_loader
    data_loader = DataLoader(
        dataset,
        batch_size=batch_size,
        shuffle=True,
        num_workers=num_workers,
        collate_fn=collate_fn,
    )

    # Initialise model
    if load_pretrained_model is not None:
        model = load_annotation_model(load_pretrained_model)
    else:
        model = get_model_instance_segmentation(2)
    model.to(device)

    # Initialise optimiser and lr_scheduler
    params = [p for p in model.parameters() if p.requires_grad]
    optimiser = torch.optim.SGD(params, lr=0.005, momentum=0.9, weight_decay=0.0005)
    lr_scheduler = torch.optim.lr_scheduler.StepLR(optimiser, step_size=3, gamma=0.1)

    # Train the model
    for epoch in range(num_epochs):
        # train for one epoch, printing every 10 iterations
        train_one_epoch(model, optimiser, data_loader, device, epoch, print_freq=10)
        # update the learning rate
        lr_scheduler.step()

        if save_intermediate or epoch == num_epochs - 1:
            save_path = outdir / f"{model_name}_{epoch}_model.pth"
            torch.save(model.state_dict(), save_path)

    print(f"Training complete. Model saved at {save_path}")
    return save_path


class Annotator(BaseModel):
    """Provides methods for automatically annotating images of flying insects using a
    pre-trained instance segmentation model.

    Parameters
    ----------
    dataset : CamfiDataset
        Dataset to annotate.
    model : Union[str, Path, MaskRCNN]
        Either a path to state dict file which defines the segmentation model, or a url
        pointing to a model to download, or one of the model names defined in
        camfi.models.model_urls.
        Alternatively, a MaskRCNN instance can be given directly.
    device : Union[str, torch.device]
        Specifies device to run inference on. E.g. set to "cuda" to use an Nvidia GPU.
    backup_device : Optional[Union[str, torch.device]]
        Specifies device to run inference on when a runtime error occurs while using
        device. Probably only makes sense to set this to "cpu" if device="cuda". This
        option enables the annotator to leverage a GPU with limited memory capacity
        without crashing if a difficult image is encountered.
    backup_model: Optional[MaskRCNN]
        Defines the backup model. Will be automatically generated if backup_device is
        set. Should not be set manually.
    split_angle : PositiveFloat
        Approximate maximum angle between polyline segments in degrees. Note that this
        will immediately be converted to radians upon instantiation of Annotator.
    poly_order : PositiveInt
        Order of polynomial used for fitting motion blur paths.
    endpoint_method : Callable[[np.ndarray, ...], tuple[NonNegativeInt, NonNegativeInt]]
        Method to find endpoints of motion blurs. The first argument to this method
        should be a cropped mask np.ndarray.
    endpoint_extra_args : list[Any]
        Extra arguments to pass to endpoint_method.
    score_thresh : float
        Score threshold between 0.0 and 1.0 for automatic annotations to be kept.
    overlap_thresh : float
        Minimum proportion of overlap (weighted intersection over minimum) between two
        instance segmentation masks to infer that one of the masks should be discarded.
    edge_thresh : NonNegativeInt
        Minimum distance an annotation has to be from the edge of the image before it is
        converted from a polyline annotation to a circle annotation.
    """

    dataset: CamfiDataset
    model: MaskRCNN = "release"
    device: Union[str, torch.device] = "cpu"
    backup_model: Optional[MaskRCNN] = None
    backup_device: Optional[Union[str, torch.device]] = None
    split_angle: PositiveFloat = 15.0
    poly_order: PositiveInt = 2
    endpoint_method: Callable[
        ..., tuple[NonNegativeInt, NonNegativeInt]
    ] = endpoint_truncate
    endpoint_extra_args: list[Any] = [10]
    score_thresh: float = 0.4
    overlap_thresh: float = 0.4
    edge_thresh: NonNegativeInt = 20

    class Config:
        arbitrary_types_allowed = True

    @validator("model", pre=True, always=True)
    def get_model(cls, v):
        if isinstance(v, MaskRCNN):
            return v
        else:
            return load_annotation_model(v)

    @validator("device", always=True)
    def put_model_on_device_and_set_to_eval(cls, v, values):
        print(f"Putting model on device: {v}", file=stderr)
        v = torch.device(v)
        values["model"].to(v)
        values["model"].eval()
        return v

    @validator("backup_model", pre=True, always=True)
    def copy_model_to_backup_device(cls, v, values):
        assert v is None, "Should not set 'backup_model'. It will be set autoimatically"
        if "backup_device" in values and values["backup_device"] is not None:
            v = copy_annotation_model(values["model"])
            v.to(values["backup_device"])
            v.eval()

        return v

    @validator("split_angle", always=True)
    def convert_split_angle_to_radians(cls, v):
        return v * pi / 180.0

    def get_prediction(self, img_idx: NonNegativeInt) -> Prediction:
        """Run predicion on a single image. First tries to use the model on self.device,
        and falls back to the model on self.backup_device if a RuntimeError is caught
        (if set).

        Parameters
        ----------
        img_idx: int
            Index of image in via project.

        Returns
        -------
        prediction: Prediction
            Output of model prediction.
        """
        try:
            img, _ = self.dataset[img_idx]
        except (OSError, RuntimeError) as e:
            print(
                f"Error loading {self.dataset.metadata(img_idx).filename}. {e!r}. Skipping.",
                file=stderr,
            )
            return Prediction.empty()

        with torch.no_grad():
            try:
                prediction = self.model([img.to(self.device)])[0]
            except RuntimeError:
                if self.backup_model:
                    prediction = self.backup_model([img.to(self.backup_device)])[0]
                else:
                    raise

        return Prediction.from_tensor_dict(prediction)

    def filter_annotations(self, prediction: Prediction) -> Prediction:
        """Applies self.score_thresh and self.overlap_thresh to filter out poor quality
        annotations.

        Parameters
        ----------
        prediction : Prediction
            Output of model prediction.

        Returns
        -------
        filtered_prediction : Prediction
            Filtered prediction.
        """
        # Remove predictions with below-threshold score
        prediction = prediction.filter_by_score(self.score_thresh)
        n_predictions = len(prediction)

        if n_predictions == 0:
            return prediction

        # Calculate mask overlaps for all pairs of predicted instances
        mask_overlaps = np.zeros((n_predictions, n_predictions), dtype="f4")

        for i, j in itertools.combinations(range(n_predictions), 2):
            if prediction.boxes[i].overlaps(prediction.boxes[j]):
                mask_overlaps[i, j] = weighted_intersection_over_minimum(
                    prediction.masks[i], prediction.masks[j]
                )
                mask_overlaps[j, i] = mask_overlaps[i, j]

        # Remove worst overlapping instances until there are no above-threshold overlaps
        keep = set(range(n_predictions))

        overlap_mask = mask_overlaps.max(axis=1) >= self.overlap_thresh
        while np.any(overlap_mask):
            # Figure out which overlapping annotation has the worst score
            overlap_annotations = np.where(overlap_mask)[0]
            to_discard = overlap_annotations[
                np.argmin(np.array(prediction.scores)[overlap_annotations])
            ]
            # Remove the annotation
            keep.remove(to_discard)
            mask_overlaps[to_discard, :] = 0.0
            mask_overlaps[:, to_discard] = 0.0

            overlap_mask = mask_overlaps.max(axis=1) >= self.overlap_thresh

        return prediction.get_subset_from_index(list(keep))

    def fit_poly(
        self,
        box: BoundingBox,
        mask: torch.Tensor,
    ) -> PolylineShapeAttributes:
        """Uses polynomial regression to fit a polyline annotation to the provided
        segmentation mask.

        Parameters
        ----------
        box : BoundingBox
            Fully contains the object to be annotated.
        mask : tensor or array
            Segmentation mask of instance with shape (image_width, image_height).

        Returns
        -------
        polyline : PolylineShapeAttributes
            Geometry of automatic annotation.
        """
        portrait = box.is_portrait()
        crop_mask = box.crop_image(mask).numpy()

        y, x = np.where(crop_mask > 0.0)
        weights = np.array(crop_mask[y, x]).flatten()

        # Set longest axis as independent variable and fit polynomial
        ind = (x, y)[portrait]
        dep = (y, x)[portrait]
        poly_fit = np.polynomial.Polynomial.fit(ind, dep, self.poly_order, w=weights)

        # Find endpoints
        ind_vals = np.arange(crop_mask.shape[not portrait])
        dep_vals = poly_fit(ind_vals)
        val_mask = np.logical_and(dep_vals < crop_mask.shape[portrait], dep_vals >= 0)
        y_vals = (dep_vals, ind_vals)[portrait][val_mask]
        x_vals = (ind_vals, dep_vals)[portrait][val_mask]
        fit_mask_vals = crop_mask[y_vals, x_vals]

        endpoints = ind_vals[
            list(self.endpoint_method(fit_mask_vals, *self.endpoint_extra_args))
        ]

        # Approximate polynomial segment with polyline
        end_gradients = poly_fit.deriv()(endpoints)
        end_angles = np.arctan(end_gradients)
        angle_diff = abs(end_angles[1] - end_angles[0])
        all_points_ind, all_points_dep = poly_fit.linspace(
            n=int(np.ceil(angle_diff / self.split_angle) + 2), domain=endpoints
        )
        polyline = PolylineShapeAttributes(
            all_points_x=list((all_points_ind, all_points_dep)[portrait] + box.x0),
            all_points_y=list((all_points_dep, all_points_ind)[portrait] + box.y0),
        )

        return polyline

    def convert_to_circle(
        self,
        polyline: PolylineShapeAttributes,
        img_shape: tuple[PositiveInt, PositiveInt],
    ) -> Union[PolylineShapeAttributes, CircleShapeAttributes]:
        """Checks if a polyline annotation is close to the edge of an image, and if so,
        converts it to a circle annotation by computing the smallest enclosing circle of
        all points in the polyline.

        Parameters
        ----------
        polyline : PolylineShapeAttributes
            Shape to convert if too close to edge.
        img_shape: tuple[int, int]
            Height and width of image.

        Returns
        -------
        shape_attributes : Union[PolylineShapeAttributes, CircleShapeAttributes]
            Geometry of annotation after (possible) conversion. If polyline does not
            go too close to the edge of the image, then polyline is returned unchanged.
            Else, a circle annotation is returned.
        """
        polyline_accepted_region = BoundingBox.from_shape(
            img_shape, border=self.edge_thresh
        )
        if polyline.in_box(polyline_accepted_region):
            return polyline

        return polyline.as_circle()

    def annotate_img(self, img_idx: int) -> list[ViaRegion]:
        """Calls self.get_prediction, self.filter_annotations, and self.fit_poly to
        produce annotations for an image specified with img_idx.

        Parameters
        ----------
        img_idx: int
            Index of image in via project.

        Returns
        -------
        regions : list[ViaRegion]
            list of annotations for image.
        """
        prediction = self.get_prediction(img_idx)
        prediction = self.filter_annotations(prediction)

        regions = []

        for i in range(len(prediction)):
            box = prediction.boxes[i]
            mask = prediction.masks[i]
            score = prediction.scores[i]
            shape_attributes: Union[CircleShapeAttributes, PolylineShapeAttributes]
            try:
                shape_attributes = self.fit_poly(box, mask)
            except (ValueError, IndexError):
                # If there is a problem fitting a polyline, use circle
                shape_attributes = CircleShapeAttributes.from_bounding_box(box)
            else:
                shape_attributes = self.convert_to_circle(
                    shape_attributes, (mask.shape[-2], mask.shape[-1])
                )
            region_attributes = ViaRegionAttributes(score=score)
            regions.append(
                ViaRegion(
                    region_attributes=region_attributes,
                    shape_attributes=shape_attributes,
                )
            )

        return regions

    def annotate(self, disable_progress_bar: Optional[bool] = True) -> ViaProject:
        """Calls self.annotate_img on all images and returns a ViaProject instance.
        Copies the `via_attributes` and `via_settings` fields from
        `self.dataset.via_project`, and just replaces the `via_img_metadata` field.

        Parameters
        ----------
        disable_progress_bar : Optional[bool]
            If True (default), progress bar is disabled.
            If set to None, disable on non-TTY.

        Returns
        -------
        project : ViaProject
            With automatic annotations made.
        """
        via_img_metadata: dict[str, ViaMetadata] = {}

        postfix = {"tot_annotations": 0}
        pb = trange(
            len(self.dataset),
            disable=disable_progress_bar,
            desc="Annotating images",
            unit="img",
            dynamic_ncols=True,
            ascii=True,
            color="green",
            postfix=postfix,
        )
        for img_idx in pb:
            img_key = self.dataset.keys[img_idx]
            regions = self.annotate_img(img_idx)
            in_metadata = self.dataset.metadata(img_idx)
            out_metadata = ViaMetadata.construct(
                file_attributes=in_metadata.file_attributes.copy(),
                filename=in_metadata.filename,
                regions=regions,
                size=in_metadata.size,
            )
            via_img_metadata[img_key] = out_metadata
            postfix["tot_annotations"] += len(regions)
            pb.set_postfix(postfix, refresh=False)

        print(f"Annotation complete.", file=stderr)
        return ViaProject(
            via_attributes=self.dataset.via_project.via_attributes,
            via_img_metadata=via_img_metadata,
            via_settings=self.dataset.via_project.via_settings,
        )


class AnnotationValidationResult(BaseModel):
    """Contains various metrics for assessing the quality of a set of automatically
    obtained annotations of flying insects.

    Parameters
    ----------
    ious : list[tuple[NonNegativeFloat, NonNegativeFloat]]
        list of (iou, score) pairs.
        iou is the Intersection over Union of the bounding boxes of true positives
        to their matched ground truth annotation. All matched annotations are
        included.
    polyline_hausdorff_distances : list[tuple[NonNegativeFloat, NonNegativeFloat]]
        list of (h_dist, score) pairs.
        h_dist is the hausdorff distance of a true positive polyline annotation,
        where the annotation is matched to a polyline ground truth annotation. Only
        polyline annotations which matched to a polyline ground truth annotation are
        included.
    length_differences : list[tuple[float, NonNegativeFloat]]
        list of (l_diff, score) pairs.
        l_diff is calculated as the length of a true positive polyline annotation
        minus the length of it's matched ground truth annotation. Only polyline
        annotations which matched to a polyline ground truth annotation are
        included.
    true_positives : list[NonNegativeFloat]
        list of scores.
    false_positives : list[NonNegativeFloat]
        list of scores. Score is the prediction score of the automatic annotation.
    false_negatives : int
        Number of false negative annotations.
    """

    ious: list[tuple[NonNegativeFloat, NonNegativeFloat]] = []
    polyline_hausdorff_distances: list[tuple[NonNegativeFloat, NonNegativeFloat]] = []
    length_differences: list[tuple[float, NonNegativeFloat]] = []
    true_positives: list[NonNegativeFloat] = []
    false_positives: list[NonNegativeFloat] = []
    false_negatives: NonNegativeInt = 0


def validate_annotations(
    auto_annotations: ViaProject,
    ground_truth: ViaProject,
    iou_thresh: float = 0.5,
    subset_functions: Optional[dict[str, Callable[[ViaMetadata], bool]]] = None,
    disable_progress_bar: Optional[bool] = True,
) -> list[AnnotationValidationResult]:
    """Compares automatic annotations against a ground-truth annotations for validation
    puposes. Validation data is stored in an AnnotationValidationResult object.

    Parameters
    ----------
    auto_annotations : ViaProject
        Automatically obtained annotations to assess.
    ground_truth : ViaProject
        Manually created ground-truth annotations.
    iou_thresh : float
        Threshold of intersection-over-union of bounding boxes to be considered a
        match. Typically, this is 0.5.
    subset_functions : Optional[dict[str, Callable[[ViaMetadata], bool]]]
        Mapping from subset name to subset function. If set, validation will be repeated
        multiple times with different subsets, once for each element.
    disable_progress_bar : Optional[bool]
        If True (default), progress bar is disabled.
        If set to None, disable on non-TTY.

    Returns
    -------
    validation_results : list[AnnotationValidationResult]
        list containing instances of AnnotationValidationResult. If subset_functions is
        set, then validation_results will have len(subset_functions) elements. By
        default it will just contain one element.
    """
    if subset_functions is None:
        subset_functions = {"all": lambda x: True}

    results: list[AnnotationValidationResult] = []

    for name, subset_function in subset_functions.items():
        annotations = auto_annotations.filtered_copy(subset_function)

        result = AnnotationValidationResult()

        for img_key, metadata in tqdm(
            annotations.via_img_metadata.items(),
            disable=disable_progress_bar,
            desc=f"Validating {name} annotations",
            unit="img",
            dynamic_ncols=True,
            ascii=True,
            color="green",
        ):
            gt_metadata = ground_truth.via_img_metadata[img_key]
            ious = sparse.dok_matrix(
                (len(metadata.regions), len(gt_metadata.regions)), dtype="f8"
            )

            for i, j in itertools.product(
                range(len(metadata.regions)), range(len(gt_metadata.regions))
            ):
                iou = metadata.regions[i].shape_attributes.interesection_over_union(
                    gt_metadata.regions[i].shape_attributes
                )
                if iou >= iou_thresh:
                    ious[i, j] = iou

            ious = ious.tocsr()
            matches = sparse.csgraph.maximum_bipartite_matching(ious, "column")
            result.false_negatives += len(gt_metadata.regions) - np.count_nonzero(
                matches >= 0
            )

            for i, match in enumerate(matches):
                score = metadata.regions[i].region_attributes.score
                if match >= 0:
                    result.true_positives.append(score)
                    result.ious.append((ious[i, match], score))
                    shape = metadata.regions[i].shape_attributes
                    gt_shape = gt_metadata.regions[match].shape_attributes
                    if shape.name == gt_shape.name == "polyline":
                        h_dist = shape.hausdorff_distance(gt_shape)
                        result.polyline_hausdorff_distances.append((h_dist, score))
                        l_diff = shape.length() - gt_shape.length()
                        result.length_differences.append((l_diff, score))

                else:
                    result.false_positives.append(score)

        results.append(result)

    return results



"""This subpackage defines the data structures used throughout camfi. Each module makes
heavy use of Pydantic's BaseModel. It is recommended to familiarise yourself with
pydantic before using this library.
"""



"""Defines data structures used during automatic annotation model training and
inference. Depends on camfi.util, camfi.datamodel.geometry, and camfi.datamodel.via."""

from __future__ import annotations

from abc import ABC, abstractmethod
from math import inf
from pathlib import Path
from typing import Optional, Union

from numpy import array
from pydantic import (
    BaseModel,
    NonNegativeFloat,
    NonNegativeInt,
    PositiveInt,
    root_validator,
    validator,
)
from skimage import draw
import torch
from torch.utils.data import Dataset

from camfi.util import dilate_idx, Field
from .geometry import (
    BoundingBox,
    PointShapeAttributes,
    CircleShapeAttributes,
    PolylineShapeAttributes,
)
from .via import ViaMetadata, ViaProject


class MaskMaker(BaseModel):
    """Defines methods to generate target instance segmentation masks for training a
    camfi annotation model.

    Parameters
    ----------
    shape : tuple[PositiveInt, PositiveInt]
        Shape of image (height, width). This will also be the shape of the instance
        segmentation masks.
    mask_dilate : Optional[PositiveInt] = None
        sets the amount of morphological dilation to apply to segmentation skeletons
        to produce instance segmentation masks.
    """

    shape: tuple[PositiveInt, PositiveInt] = Field(
        ..., description="Shape of images (height, width) in pixels."
    )
    mask_dilate: Optional[PositiveInt] = Field(
        None, description="Morphological dilation to apply to segmentation skeletons."
    )

    class Config:
        schema_extra = {
            "description": "Contains settings for segmentation mask generation."
        }

    def get_point_mask(self, point: PointShapeAttributes) -> torch.Tensor:
        """Produces a mask Tensor for a given point. Raises a ValueError if point lies
        outside self.shape.

        Parameters
        ----------
        point : PointShapeAttributes
            Geometry from which to generate instance segmentation mask.

        Returns
        -------
        mask : torch.Tensor
            Instance segmentation mask.
        """
        cx = array([int(point.cx)])
        cy = array([int(point.cy)])

        if cx >= self.shape[1] or cy >= self.shape[0]:
            raise ValueError(
                f"point ({cy}, {cx}) lies outside image with shape {self.shape}"
            )

        mask = torch.zeros(self.shape)
        if self.mask_dilate is None:
            rr, cc = cy, cx
        else:
            rr, cc = dilate_idx(cy, cx, self.mask_dilate, img_shape=self.shape)

        mask[rr, cc] = 1

        return mask

    def get_circle_mask(self, circle: CircleShapeAttributes) -> torch.Tensor:
        """Produces a mask Tensor for a given circle. Raises a ValueError if centre lies
        outside self.shape.

        Parameters
        ----------
        circle : CirlceShapeAttributes
            Geometry from which to generate instance segmentation mask.

        Returns
        -------
        mask : torch.Tensor
            Instance segmentation mask.
        """
        return self.get_point_mask(circle.as_point())

    def get_polyline_mask(self, polyline: PolylineShapeAttributes) -> torch.Tensor:
        """Produces a mask Tensor for a given polyline. Raises a ValueError if any point
        lies outside self.shape.

        Parameters
        ----------
        polyline : PolylineShapeAttributes
            Geometry from which to generate instance segmentation mask.

        Returns
        -------
        mask : torch.Tensor
            Instance segmentation mask.
        """
        x = polyline.all_points_x
        y = polyline.all_points_y
        mask = torch.zeros(self.shape)

        for i in range(len(x) - 1):
            if x[i] >= self.shape[1] or y[i] >= self.shape[0]:
                raise ValueError(
                    f"point ({x[i]}, {y[i]}) lies outside image with shape {self.shape}"
                )

            rr, cc = draw.line(int(y[i]), int(x[i]), int(y[i + 1]), int(x[i + 1]))
            if self.mask_dilate is not None:
                rr, cc = dilate_idx(rr, cc, self.mask_dilate, img_shape=self.shape)

            mask[rr, cc] = 1

        return mask

    def get_mask(
        self,
        shape_attributes: Union[
            PointShapeAttributes, CircleShapeAttributes, PolylineShapeAttributes
        ],
    ) -> torch.Tensor:
        """Produces a mask Tensor for a given geometry. Raises a ValueError if any point
        lies outside self.shape. Calls one of self.get_point_mask,
        self.get_circle_mask, or self.get_polyline_mask, depending on the type of
        shape_attributes.

        Parameters
        ----------
        shape_attributes : Union[
            PointShapeAttributes, CircleShapeAttributes, PolylineShapeAttributes
        ]
            Geometry from which to generate instance segmentation mask.

        Returns
        -------
        mask : torch.Tensor
            Instance segmentation mask.
        """
        if isinstance(shape_attributes, PointShapeAttributes):
            return self.get_point_mask(shape_attributes)
        elif isinstance(shape_attributes, CircleShapeAttributes):
            return self.get_circle_mask(shape_attributes)
        else:
            return self.get_polyline_mask(shape_attributes)

    def get_masks(self, metadata: ViaMetadata) -> list[torch.Tensor]:
        """Calls self.get_mask on all regions in metadata.

        Parameters
        ----------
        metadata : ViaMetadata
            Annotation data for an image.

        Returns
        -------
        masks : list[torch.Tensor]
            list of instance segmentation masks. One for each item in
            metadata.regions.
        """
        return [self.get_mask(region.shape_attributes) for region in metadata.regions]


class TargetPredictionABC(BaseModel, ABC):
    """Abstract base class for defining Targets (the input to the instance segmentation
    model training) and Predictions (the ouput of instance segmentation inference).

    Parameters
    ----------
    boxes : list[BoundingBox]
        list of bounding boxes of annotated objects in an image.
    labels : list[PositiveInt]
        list of integer class labels of annotated objects in an image. Zero is reserved
        for the background.
    masks : list[torch.Tensor]
        list of instance segmentation masks of annotated objects in an image.
    """

    boxes: list[BoundingBox]
    labels: list[PositiveInt]
    masks: list[torch.Tensor]

    class Config:
        """Pydantic configuration for TargetPredictionABC."""

        arbitrary_types_allowed = True

    @root_validator
    def all_fields_have_same_length(cls, values):
        """Pydantic validation method, called when instantiating subclasses of
        TargetPredictionABC. Ensures that all fields have the same length."""
        try:
            length = len(values["boxes"])
            if not all(len(values[k]) == length for k in ["labels", "masks"]):
                raise ValueError("Fields must have same length")
        except KeyError:
            raise ValueError("Invalid parameters given to Target")

        return values

    @validator("masks")
    def all_masks_same_shape(cls, v):
        """Pydantic validation method, called when instantiating subclasses of
        TargetPredictionABC. Ensures that all masks have the same shape."""
        if len(v) <= 1:
            return v

        shape = v[0].shape
        for mask in v[1:]:
            if mask.shape != shape:
                raise ValueError("All masks must have same shape")

        return v

    def __len__(self):
        return len(self.labels)

    @abstractmethod
    def to_tensor_dict(self) -> dict[str, torch.Tensor]:
        """Send data to a dict of Tensors.

        Returns
        -------
        tensor_dict : dict[str, torch.Tensor]
            Compatible with pytorch instance segmentation model.
        """

    @classmethod
    @abstractmethod
    def from_tensor_dict(
        cls, tensor_dict: dict[str, torch.Tensor]
    ) -> TargetPredictionABC:
        """Load Target or Prediction from tensor_dict.

        Parameters
        -------
        tensor_dict : dict[str, torch.Tensor]
            Compatible with pytorch instance segmentation model.

        Returns
        -------
        target_or_prediction : TargetPredictionABC
            Sub-classes of TargetPredictionABC should define this to return an instance
            of themselves.
        """


class Target(TargetPredictionABC):
    """Sub-class of TargetPredictionABC, defining the input to camfi model training.

    Parameters
    ----------
    boxes : list[BoundingBox]
        list of bounding boxes of annotated objects in an image.
    labels : list[PositiveInt]
        list of integer class labels of annotated objects in an image. Zero is reserved
        for the background.
    masks : list[torch.Tensor]
        list of instance segmentation masks of annotated objects in an image.
    image_id : NonNegativeInt
        Unique index of image to which Target instance relates.
    """

    image_id: NonNegativeInt

    def to_tensor_dict(self) -> dict[str, torch.Tensor]:
        """Send data to a dict of Tensors.

        Returns
        -------
        tensor_dict : dict[str, torch.Tensor]
            Compatible with pytorch instance segmentation model.
        """
        return dict(
            boxes=torch.tensor([[b.x0, b.y0, b.x1, b.y1] for b in self.boxes]),
            labels=torch.tensor(self.labels),
            image_id=torch.tensor([self.image_id]),
            masks=torch.stack(self.masks),
        )

    @classmethod
    def from_tensor_dict(cls, tensor_dict: dict[str, torch.Tensor]) -> Target:
        """Load Target from tensor_dict.

        Parameters
        -------
        tensor_dict : dict[str, torch.Tensor]
            Compatible with pytorch instance segmentation model.

        Returns
        -------
        target : Target
            For use by camfi.
        """
        return Target(
            boxes=[
                BoundingBox(x0=x0, y0=y0, x1=x1, y1=y1)
                for x0, y0, x1, y1 in tensor_dict["boxes"]
            ],
            labels=[int(v) for v in tensor_dict["labels"]],
            masks=list(tensor_dict["masks"]),
            image_id=int(tensor_dict["image_id"]),
        )

    @classmethod
    def empty(cls, image_id: NonNegativeInt = 0) -> Target:
        """Initialises a Target with and image_id but no other data.

        Parameters
        ----------
        image_id : NonNegativeInt
            Unique image index to initialise Target with.

        Returns
        -------
        target : Target
            Target instance with no data (except image_id).

        Examples
        --------
        >>> Target.empty()
        Target(boxes=[], labels=[], masks=[], image_id=0)
        """
        return Target.construct(boxes=[], labels=[], image_id=image_id, masks=[])


class Prediction(TargetPredictionABC):
    """Sub-class of TargetPredictionABC, defining the output of camfi model inference.

    Parameters
    ----------
    boxes : list[BoundingBox]
        list of bounding boxes of annotated objects in an image.
    labels : list[PositiveInt]
        list of integer class labels of annotated objects in an image. Zero is reserved
        for the background.
    masks : list[torch.Tensor]
        list of instance segmentation masks of annotated objects in an image.
    scores : list[NonNegativeFloat]
        list of annotation scores (from 0.0 to 1.0)
    """

    scores: list[NonNegativeFloat]

    @root_validator
    def all_fields_have_same_length(cls, values):
        """Pydantic validation method, called when instantiating Prediction.
        Ensures that all fields have the same length."""
        length = len(values["boxes"])
        if not all(len(values[k]) == length for k in ["labels", "masks", "scores"]):
            raise ValueError("Fields must have same length")

        return values

    def to_tensor_dict(self) -> dict[str, torch.Tensor]:
        """Send data to a dict of Tensors.

        Returns
        -------
        tensor_dict : dict[str, torch.Tensor]
            Compatible with pytorch instance segmentation model.
        """
        return dict(
            boxes=torch.tensor([[b.x0, b.y0, b.x1, b.y1] for b in self.boxes]),
            labels=torch.tensor(self.labels),
            scores=torch.tensor(self.scores),
            masks=torch.stack(self.masks),
        )

    @classmethod
    def from_tensor_dict(cls, tensor_dict: dict[str, torch.Tensor]) -> Prediction:
        """Load Prediction from tensor_dict.

        Parameters
        -------
        tensor_dict : dict[str, torch.Tensor]
            Compatible with pytorch instance segmentation model.

        Returns
        -------
        prediction : Prediction
            For use by camfi.
        """
        return Prediction(
            boxes=[
                BoundingBox(x0=x0, y0=y0, x1=x1, y1=y1)
                for x0, y0, x1, y1 in tensor_dict["boxes"]
            ],
            labels=[int(v) for v in tensor_dict["labels"]],
            masks=list(tensor_dict["masks"]),
            scores=[int(v) for v in tensor_dict["scores"]],
        )

    @classmethod
    def empty(cls) -> Prediction:
        """Initialises an Prediction with no data.

        Returns
        -------
        prediction : Prediction
            Prediction instance with no data.

        Examples
        --------
        >>> Prediction.empty()
        Prediction(boxes=[], labels=[], masks=[], scores=[])
        """
        return Prediction.construct(boxes=[], labels=[], masks=[], scores=[])

    def filter_by_score(self, score_thresh: float) -> Prediction:
        """Returns a Prediction instance with items with scores below score_thresh
        removed.

        Parameters
        ----------
        score_thresh: float
            Score threshold. Annotations with score below score_thresh are not
            included in the returned Prediction object.

        Returns
        -------
        prediction : Prediction
            Prediction object with below-score-threshold annotations removed.
        """
        filtered_prediction = Prediction.empty()
        for i in range(len(self)):
            if self.scores[i] >= score_thresh:
                filtered_prediction.boxes.append(self.boxes[i])
                filtered_prediction.labels.append(self.labels[i])
                filtered_prediction.masks.append(self.masks[i])
                filtered_prediction.scores.append(self.scores[i])

        return filtered_prediction

    def get_subset_from_index(self, subset: list[NonNegativeInt]) -> Prediction:
        """Returns a Prediction instance from self with items indexed by the elements
        of subset.

        Parameters
        ----------
        subset : list[NonNegativeInt]
            list of indices to include in subset.

        Returns
        -------
        prediction : Prediction
            Prediction object only including specified subset of annotations.

        Examples
        --------
        >>> prediction = Prediction(
        ...     boxes=[
        ...         BoundingBox(x0=0, y0=0, x1=1, y1=1),
        ...         BoundingBox(x0=1, y0=1, x1=2, y1=2),
        ...     ],
        ...     labels=[1, 2],
        ...     masks=[torch.zeros(2, 2), torch.zeros(2, 2)],
        ...     scores=[0.0, 1.0],
        ... )
        >>> subset_prediction = prediction.get_subset_from_index([0])
        >>> subset_prediction.boxes == prediction.boxes[0:1]
        True
        >>> subset_prediction.labels == prediction.labels[0:1]
        True
        >>> len(subset_prediction.masks) == 1
        True
        >>> subset_prediction.scores == prediction.scores[0:1]
        True
        """
        filtered_prediction = Prediction.empty()
        for i in subset:
            filtered_prediction.boxes.append(self.boxes[i])
            filtered_prediction.labels.append(self.labels[i])
            filtered_prediction.masks.append(self.masks[i])
            filtered_prediction.scores.append(self.scores[i])

        return filtered_prediction


class ImageTransform(BaseModel, ABC):
    """Abstract base class for transforms on images with segmentation data.
    Instances of subclasses of ImageTransform are callables which when called, apply
    a transformation to an image Tensor and an accosiated target_dict.
    """

    def __call__(
        self, image: torch.Tensor, target: Target
    ) -> tuple[torch.Tensor, Target]:
        image, target_dict = self.apply_to_tensor_dict(image, target.to_tensor_dict())
        return image, Target.from_tensor_dict(target_dict)

    @abstractmethod
    def apply_to_tensor_dict(
        self, image: torch.Tensor, target: dict[str, torch.Tensor]
    ) -> tuple[torch.Tensor, dict[str, torch.Tensor]]:
        """Subclasses of ImageTransform should implement this to return a transformed
        image and target dict.

        Parameters
        ----------
        image : torch.Tensor
            Image tensor on which annotations are made.
        target : dict[str, torch.Tensor]
            Specifies annotations on image. Output of Target.to_tensor_dict().

        Returns
        -------
        transformed_image : torch.Tensor
            Transformed image.
        transformed_target : dict[str, torch.Tensor]
            Specifies transformed annotations on image. Can be used as input to
            Target.from_tensor_dict().
        """


class CamfiDataset(BaseModel, Dataset):
    """Defines a camfi dataset, compatible with torch.utils.data.DataLoader, for
    loading image (and annotation) data for camfi model training and inference.

    When indexed with an int, a CamfiDataset object returns an image Tensor and a Target
    object, which contains the manual annotations for that image. If using in inference
    mode, the Target will not contain any data (except the image index).

    Parameters
    ----------
    root: Path
        Path to directory containing images (in subdirectories). Filenames specified in
        via_project should be relative paths from this root directory.
    via_project: ViaProject
        Via project data, containing all annotation data, and all images to be trained
        on (or for inference to run on).
    crop: Optional[BoundingBox]
        If specified, images will be cropped to this bounding box after loading them
        from file. This can be useful when working with images taken by trail cameras,
        as some models burn a information bar onto the bottom of the image, which
        ideally should be removed before training or inference on that image.
    inference_mode: bool
        Flag to indicate whether dataset is being used for inference or training. In
        training mode, data augmentation etc. will be performed on the dataset.
    mask_maker: Optional[MaskMaker]
        Instance of MaskMaker which produces instance segementation masks for model
        training. Only set if inference_mode = False.
    transform: Optional[ImageTransform]
        Image transformations to apply for data augmentation. Only set
        if inference_mode = False.
    min_annotations: int
        Only train on images which have at least this many annotations.
        Only set if inference_mode = False.
    max_annotations: float = inf
        Only train on images which have at most this many annotations. Can be useful
        if running into memory errors on the GPU, since memory consumption depends on
        the number of annotated objects in the image.
        Only set if inference_mode = False.
    box_margin: PositiveInt = 10
        Margin to add to bounding boxes of object annotations, for model training.
        Only set if inference_mode = False.
    exclude: set[Path]
        Optionally specify a set of image files to exclude for training or inference.
        Elements in exclude must match the filenames (ie. relative file paths) of
        images to exclude.
    keys: list[str]
        list of image keys. This field is automatically generated upon instantiation,
        and should not be set.
    """

    root: Path
    via_project: ViaProject
    crop: Optional[BoundingBox] = None

    inference_mode: bool = False

    # Only set if inference_mode = False
    mask_maker: Optional[MaskMaker] = None
    transform: Optional[ImageTransform] = None
    min_annotations: Optional[int] = None
    max_annotations: Optional[int] = None
    box_margin: PositiveInt = 10

    # Optionally exclude some files
    exclude: set[Path] = None  # type: ignore[assignment]

    # Automatically generated. No need to set.
    keys: list[str] = None  # type: ignore[assignment]

    @validator("exclude", pre=True, always=True)
    def default_exclude(cls, v):
        """Pydantic validation method, called when instantiating CamfiDataset.
        sets default for exclude."""
        if v is None:
            return set()
        return v

    @validator("transform")
    def only_set_transform_if_not_inference_mode(cls, v, values):
        """Pydantic validation method, called when instantiating CamfiDataset.
        Ensures that transform is only set if in training mode."""
        if "inference_mode" in values and values["inference_mode"] is True:
            assert v is None, "Only set if inference_mode=False"
        return v

    @validator("min_annotations")
    def only_set_min_annotations_if_not_inference_mode(cls, v, values):
        """Pydantic validation method, called when instantiating CamfiDataset.
        Ensures that min_annotations is only set if in training mode."""
        if "inference_mode" in values and values["inference_mode"] is True:
            assert v is None, "Only set if inference_mode=False"
        return v

    @validator("max_annotations")
    def only_set_max_annotations_if_not_inference_mode(cls, v, values):
        """Pydantic validation method, called when instantiating CamfiDataset.
        Ensures that max_annotations is only set if in training mode."""
        if "inference_mode" in values and values["inference_mode"] is True:
            assert v is None, "Only set if inference_mode=False"
        return v

    @validator("box_margin")
    def only_set_box_margin_if_not_inference_mode(cls, v, values):
        """Pydantic validation method, called when instantiating CamfiDataset.
        Ensures that box_margin is only set if in training mode."""
        if "inference_mode" in values and values["inference_mode"] is True:
            assert v == 10, "Only set if inference_mode=False"
        return v

    @validator("mask_maker", always=True)
    def set_iff_not_inference_mode(cls, v, values):
        """Pydantic validation method, called when instantiating CamfiDataset.
        Ensures that mask_maker is set iff in training mode."""
        if "inference_mode" in values and values["inference_mode"] is True:
            assert v is None, "Only set if inference_mode=False"
        else:
            assert isinstance(
                v, MaskMaker
            ), "mask_maker must be set if inference_mode=False"
        return v

    @validator("keys", pre=True, always=True)
    def generate_filtered_keys(cls, v, values):
        """Pydantic validation method, called when instantiating CamfiDataset.
        Generates list of image keys."""
        min_annotations = values.get("min_annotations", 0)
        if min_annotations is None:
            min_annotations = 0
        max_annotations = values.get("max_annotations", inf)
        if max_annotations is None:
            max_annotations = inf
        return list(
            dict(
                filter(
                    lambda e: e[1].filename not in values["exclude"]
                    and min_annotations <= len(e[1].regions) <= max_annotations,
                    values["via_project"].via_img_metadata.items(),
                )
            )
        )

    def __getitem__(self, idx: int) -> tuple[torch.Tensor, Target]:
        metadata = self.via_project.via_img_metadata[self.keys[idx]]
        image = metadata.read_image(root=self.root)

        if self.crop is not None:
            image = self.crop.crop_image(image)

        if self.inference_mode:
            target = Target.empty(image_id=idx)
        else:  # Training mode
            boxes = metadata.get_bounding_boxes()
            for box in boxes:
                box.add_margin(self.box_margin)
            target = Target(
                boxes=boxes,
                labels=metadata.get_labels(),
                image_id=idx,
                masks=self.mask_maker.get_masks(metadata),  # type: ignore[union-attr]
            )

            if self.transform is not None:
                image, target = self.transform(image, target)

        return image, target

    def __len__(self):
        return len(self.keys)

    def metadata(self, idx: int) -> ViaMetadata:
        """Returns the ViaMetadata object of the image at idx.

        Parameters
        ----------
        idx : int
            Index of image in dataset.

        Returns
        -------
        metadata : ViaMetadata
            Annotation data for the image.
        """
        return self.via_project.via_img_metadata[self.keys[idx]]
